//****************************************************************************
//**
//**    IN_MAIN.CPP
//**    Input - Main Interface
//**
//**    Copyright (c) 1998 Chris Hargrove
//**    Notice: Usage of any code in this file is subject to the rules
//**            described in the LICENSE.TXT file included in this directory.
//**            Reading, compiling, or otherwise using this code constitutes
//**            automatic acceptance of the rules in said text file.
//**
//****************************************************************************
//============================================================================
//    IMPLEMENTATION HEADERS
//============================================================================
#define SYS_DEFS_INCLUDE_WIN32
#include "sys_defs.h"
#include "in_main.h"
#include "in_win.h"

//============================================================================
//    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
//============================================================================
#define IN_MAXRECEIVERS 32 // maximum number of receivers

//============================================================================
//    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
//============================================================================

// Input receiver type
// Entry in a chain of callbacks that user input is fed through
// Receiver functions return true if the input has been accepted, or false
// if it should be continued down the chain.  Receivers can safely call
// the Remove function within themselves.

typedef struct in_receiver_s in_receiver_t;
struct in_receiver_s
{
	bbool (*handler)(in_event_t *);
	in_receiver_t *prev, *next;
};

//============================================================================
//    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE DATA
//============================================================================
// receiver data
static in_receiver_t in_ReceiverPool[IN_MAXRECEIVERS];
static in_receiver_t in_ReceiverChain;

// shifted key table
static dword in_Shifted[256];

// key name list
struct keyName_s
{
	char *name;
	dword key;
	char *numname;
} in_KeyNames[] =
{
	{ "BACKSPACE", INKEY_BACKSPACE, NULL},
	{ "TAB", INKEY_TAB, NULL},
	{ "ENTER", INKEY_ENTER, NULL},
	{ "ESCAPE", INKEY_ESCAPE, NULL},
	{ "SPACE", INKEY_SPACE, NULL},

	{ "LEFTSHIFT", INKEY_LEFTSHIFT, NULL},
	{ "RIGHTSHIFT", INKEY_RIGHTSHIFT, NULL},
	{ "LEFTCTRL", INKEY_LEFTCTRL, NULL},
	{ "RIGHTCTRL", INKEY_RIGHTCTRL, NULL},
	{ "LEFTALT", INKEY_LEFTALT, NULL},
	{ "RIGHTALT", INKEY_RIGHTALT, NULL},
	
	{ "LEFTARROW", INKEY_LEFTARROW, NULL},
	{ "RIGHTARROW", INKEY_RIGHTARROW, NULL},
	{ "UPARROW", INKEY_UPARROW, NULL},
	{ "DOWNARROW", INKEY_DOWNARROW, NULL},

	{ "F1", INKEY_F1, NULL},
	{ "F2", INKEY_F2, NULL},
	{ "F3", INKEY_F3, NULL},
	{ "F4", INKEY_F4, NULL},
	{ "F5", INKEY_F5, NULL},
	{ "F6", INKEY_F6, NULL},
	{ "F7", INKEY_F7, NULL},
	{ "F8", INKEY_F8, NULL},
	{ "F9", INKEY_F9, NULL},
	{ "F10", INKEY_F10, NULL},
	{ "F11", INKEY_F11, NULL},
	{ "F12", INKEY_F12, NULL},

	{ "INS", INKEY_INS, NULL},
	{ "DEL", INKEY_DEL, NULL},
	{ "HOME", INKEY_HOME, NULL},
	{ "END", INKEY_END, NULL},
	{ "PGUP", INKEY_PGUP, NULL},
	{ "PGDN", INKEY_PGDN, NULL},

	{ "NUM/", INKEY_NUMSLASH, "/"},
	{ "NUM*", INKEY_NUMSTAR, "*"},
	{ "NUM-", INKEY_NUMMINUS, "-"},
	{ "NUM+", INKEY_NUMPLUS, "+"},
	{ "NUMENTER", INKEY_NUMENTER, NULL},
	{ "NUM.", INKEY_NUMPERIOD, "."},
	{ "NUM0", INKEY_NUM0, "0"},
	{ "NUM1", INKEY_NUM1, "1"},
	{ "NUM2", INKEY_NUM2, "2"},
	{ "NUM3", INKEY_NUM3, "3"},
	{ "NUM4", INKEY_NUM4, "4"},
	{ "NUM5", INKEY_NUM5, "5"},
	{ "NUM6", INKEY_NUM6, "6"},
	{ "NUM7", INKEY_NUM7, "7"},
	{ "NUM8", INKEY_NUM8, "8"},
	{ "NUM9", INKEY_NUM9, "9"},

	{ "NUMLOCK", INKEY_NUMLOCK, NULL},
	{ "CAPSLOCK", INKEY_CAPSLOCK, NULL},
	{ "SCROLLLOCK", INKEY_SCROLLLOCK, NULL},
	{ "PRINTSCREEN", INKEY_PRINTSCRN, NULL},
	{ "PAUSE", INKEY_PAUSE, NULL},
	
	{ "MOUSE1", INKEY_MOUSELEFT, NULL},
	{ "MOUSE2", INKEY_MOUSERIGHT, NULL},
	{ "MOUSE3", INKEY_MOUSEMIDDLE, NULL},

	// give names to a few regular keys that tend to be reserved
	{ "SEMICOLON", ';', ";"},
	{ "QUOTE", '\'', "\'"},
	{ "TILDE", '`', "`"},

	{ NULL, 0, NULL }
};

//============================================================================
//    INTERFACE DATA
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
//============================================================================
static void ReceiverDispatch(in_event_t *event);

//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTIONS
//============================================================================

// event handler callback
static void ReceiverDispatch(in_event_t *event)
{
	in_receiver_t *rec, *next;

	for (rec=in_ReceiverChain.next; rec!=&in_ReceiverChain; rec=next)
	{
		next = rec->next; // use a safe next pointer incase receiver removes self
		if (rec->handler(event))
			return;
	}
}

//============================================================================
//    INTERFACE FUNCTIONS
//============================================================================

// initialization
void IN_Init()
{
	int i;
	
	INW_Init(); // init directinput
	INW_SetEventHandler(ReceiverDispatch); // set event handler to dispatcher
	
	// we're doing fullscreen stuff now, so set the mouse to exclusive mode
	INW_SetMouseExclusive(1);

	// link receiver head to itself
	in_ReceiverChain.next = in_ReceiverChain.prev = &in_ReceiverChain;

	// set up shifted key table
	// keys with shift entries of zero don't have shifted versions
	for (i=0;i<26;i++)
		in_Shifted[i+'a'] = i+'A';
	in_Shifted['0'] = ')';
	in_Shifted['1'] = '!';
	in_Shifted['2'] = '@';
	in_Shifted['3'] = '#';
	in_Shifted['4'] = '$';
	in_Shifted['5'] = '%';
	in_Shifted['6'] = '^';
	in_Shifted['7'] = '&';
	in_Shifted['8'] = '*';
	in_Shifted['9'] = '(';
	in_Shifted['-'] = '_';
	in_Shifted['='] = '+';
	in_Shifted['['] = '{';
	in_Shifted[']'] = '}';
	in_Shifted['\\'] = '|';
	in_Shifted[';'] = ':';
	in_Shifted['\''] = '"';
	in_Shifted[','] = '<';
	in_Shifted['.'] = '>';
	in_Shifted['/'] = '?';
	in_Shifted['`'] = '~';
}

// shutdown
void IN_Shutdown()
{
	INW_Shutdown(); // shut down directinput
}

// per-frame operation
void IN_Frame()
{
	INW_Frame(); // run directinput frame
}

// allocate a new receiver on the receiver stack
// returns 0 if the allocation failed, or a handle to the receiver
dword IN_AddReceiver(bbool (*handler)(in_event_t *))
{
	int i;
	in_receiver_t *rec;

	for (i=0;i<IN_MAXRECEIVERS;i++)
	{
		rec = &in_ReceiverPool[i];
		if (!rec->handler)
		{
			// receiver is available
			rec->handler = handler;
			// add it to the list
			rec->prev = &in_ReceiverChain;
			rec->next = in_ReceiverChain.next;
			rec->prev->next = rec;
			rec->next->prev = rec;
			return((dword)(rec - in_ReceiverPool) + 1);
		}
	}
	return(0);
}

// change the handler of an existing receiver
void IN_ChangeReceiverHandler(dword receiver, bbool (*handler)(in_event_t *))
{
	in_receiver_t *rec;

	if ((!receiver) || (receiver > IN_MAXRECEIVERS))
		return; // trying to free an out-of-range receiver
	rec = &in_ReceiverPool[receiver - 1];
	if (!rec->handler)
		return; // trying to free an already free receiver

	rec->handler = handler;
}

// take a receiver out of the stack and free it, takes handle as parameter
void IN_RemoveReceiver(dword receiver)
{
	in_receiver_t *rec;

	if ((!receiver) || (receiver > IN_MAXRECEIVERS))
		return; // trying to free an out-of-range receiver
	rec = &in_ReceiverPool[receiver - 1];
	if (!rec->handler)
		return; // trying to free an already free receiver
	
	rec->handler = NULL;
	// unlink from the list
	rec->prev->next = rec->next;
	rec->next->prev = rec->prev;
}

// return the string name for a key
char* IN_NameForKey(dword key)
{
	int i;
	static char tempStr[2] = {0,0};

	if ((!key) || (key >= INKEY_NUMKEYS))
		return(NULL); // out of range
	for (i=0; in_KeyNames[i].name; i++)
	{
		if (in_KeyNames[i].key == key)
		{
			return(in_KeyNames[i].name);
		}
	}
	// not an extended key, use regular characters
	if ((key < ' ') || (key > 127))
		return(NULL); // out of range
	if ((key >= 'a') && (key <= 'z'))
		key = IN_GetShiftedKey(key); // use uppercase for letters
	tempStr[0] = key;
	return(tempStr);
}

// return a key matching the string name
dword IN_KeyForName(char* name)
{
	dword i;
	char *str;

	for (i=0;i<INKEY_NUMKEYS;i++)
	{
		if ((str = IN_NameForKey(i)) && (!stricmp(name, str)))
			return(i);
	}
	return(0);
}

// returns a version of a key as if SHIFT were held down
// only works with regular characters
dword IN_GetShiftedKey(dword key)
{
	if (in_Shifted[key])
		return(in_Shifted[key]);
	return(key);
}

//============================================================================
//    INTERFACE CLASS BODIES
//============================================================================

//****************************************************************************
//**
//**    END MODULE IN_MAIN.CPP
//**
//****************************************************************************

