//****************************************************************************
//**
//**    IN_WIN.CPP
//**    Input - Windows DirectInput
//**
//**    Copyright (c) 1998 Chris Hargrove
//**    Notice: Usage of any code in this file is subject to the rules
//**            described in the LICENSE.TXT file included in this directory.
//**            Reading, compiling, or otherwise using this code constitutes
//**            automatic acceptance of the rules in said text file.
//**
//****************************************************************************
//============================================================================
//    IMPLEMENTATION HEADERS
//============================================================================
#define SYS_DEFS_INCLUDE_WIN32
#include "sys_defs.h"
#include "in_win.h"

// dinput.h defines these so undef 'em
#undef byte
#include <dinput.h>

//============================================================================
//    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
//============================================================================
#define INW_INPUTBUFFERSIZE 128 // maximum pieces of dinput data per frame
#define INW_NUMDIKEYS 256 // maximum number of directinput DIK_ keys
#define INW_MAXINJECT 128 // maximum injected key events per frame

//============================================================================
//    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
//============================================================================
//============================================================================
//    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE DATA
//============================================================================

// directinput device crap
static LPDIRECTINPUT lpDI=NULL;
static LPDIRECTINPUTDEVICE lpMouse=NULL;
static LPDIRECTINPUTDEVICE lpKeyboard=NULL;
static DIDEVCAPS DICaps;

// directinput DIK_ to INKEY_ key map, filled in at init
static dword inw_DIKToINKEY[INW_NUMDIKEYS];

// current and previous key states
static byte keyStates[INKEY_NUMKEYS];
static byte oldKeyStates[INKEY_NUMKEYS];

// time since keys were "press"ed, for key delays etc.
static float keyPressTimes[INKEY_NUMKEYS];

// mouse positions during key events for drag deltas
static sdword keyMouseDragPositions[INKEY_NUMKEYS][2]; // x/y

// mouse limits, set to "no limit" extremes by default
static sdword mouseMinX=0x80000000, mouseMinY=0x80000000; // signed minimum
static sdword mouseMaxX=0x7fffffff, mouseMaxY=0x7fffffff; // signed maximum

// current mouse positions
static sdword mouseX, mouseY; // within limits
static sdword mouseFreeX, mouseFreeY; // without limits
// note: 32 bits is sufficient for the free coordinates unless you set a
//       super-high mouse sensitivity or feel like rolling in one direction
//       for a few weeks straight with no break... uhh, i don't think so.

// mouse sensitivity multiplier
static float mouseSensitivity = 1.0;

// event callback
static void (*inw_eventHandler)(in_event_t *) = NULL;

// exclusive mode toggle
static bbool mouseExclusive = 1;

// device acquired status
static bbool mouseAcquired, keyboardAcquired;

// keyboard event injection buffer
static dword keyInjectionBuffer[INW_MAXINJECT][2]; // key/state
static dword keyInjectionBufferIndex;

// event index
static dword inw_eventIndex;

//============================================================================
//    INTERFACE DATA
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
//============================================================================
static char* NameForDIError(HRESULT err);
static HRESULT SetMouseCooperative(DWORD inFlags);
static void FlushKeyboardData();
static void InitDIKToINKEY();
static void AcquireMouse();
static void UnacquireMouse();
static void AcquireKeyboard();
static void UnacquireKeyboard();
static void CheckKeyEvents(dword key, in_event_t *event);

//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTIONS
//============================================================================
static char* NameForDIError(HRESULT err)
{
	switch(err)
	{
		case DI_OK: return("OK"); break;
		//case DI_NOTATTACHED: return("NOTATTACHED"); break;
		//case DI_PROPNOEFFECT: return("PROPNOEFFECT"); break;
		case DI_BUFFEROVERFLOW: return("NOTATTACHED/PROPNOEFFECT/BUFFEROVERFLOW"); break;
		case DIERR_INPUTLOST: return("INPUTLOST"); break;
		case DIERR_INVALIDPARAM: return("INVALIDPARAM"); break;
		//case DIERR_OTHERAPPHASPRIO: return("OTHERAPPHASPRIO"); break;
		//case DIERR_HANDLEEXISTS: return("HANDLEEXISTS"); break;
		case DIERR_READONLY: return("OTHERAPPHASPRIO/HANDLEEXISTS/READONLY"); break;
		case DIERR_ACQUIRED: return("ACQUIRED"); break;
		case DIERR_NOTACQUIRED: return("NOTACQUIRED"); break;
		case DIERR_NOAGGREGATION: return("NOAGGREGATION"); break;
		case DIERR_ALREADYINITIALIZED: return("ALREADYINITIALIZED"); break;
		case DIERR_NOTINITIALIZED: return("NOTINITIALIZED"); break;
		case DIERR_UNSUPPORTED: return("UNSUPPORTED"); break;
		case DIERR_OUTOFMEMORY: return("OUTOFMEMORY"); break;
		case DIERR_GENERIC: return("GENERIC"); break;
		case DIERR_NOINTERFACE: return("NOINTERFACE"); break;
		case DIERR_DEVICENOTREG: return("DEVICENOTREG"); break;
		case DIERR_OBJECTNOTFOUND: return("OBJECTNOTFOUND"); break;
		case DIERR_BETADIRECTINPUTVERSION: return("BETADIRECTINPUTVERSION"); break;
		case DIERR_BADDRIVERVER: return("BADDRIVERVER"); break;
		case DI_POLLEDDEVICE: return("POLLEDDEVICE"); break;
		default:
			break;
	}
	return("Unknown");
}

static HRESULT SetMouseCooperative(DWORD inFlags)
{
	if (mouseExclusive)
		inFlags |= DISCL_EXCLUSIVE;
	else
		inFlags |= DISCL_NONEXCLUSIVE;
	return(lpMouse->SetCooperativeLevel(sysw_hwnd, inFlags));
}

static void FlushKeyboardData()
{
	if (lpKeyboard)
	{
		dword num;
	
		// flush the dinput buffer, don't worry about return value
		num = INFINITE;
		lpKeyboard->GetDeviceData(sizeof(DIDEVICEOBJECTDATA),NULL,&num,0);
	}

	// wipe our internal key data
	memset(keyStates, 0, sizeof(byte)*INKEY_NUMKEYS);
	memset(oldKeyStates, 0, sizeof(byte)*INKEY_NUMKEYS);
	memset(keyPressTimes, 0, sizeof(float)*INKEY_NUMKEYS);
	keyInjectionBufferIndex = 0;
}

static void InitDIKToINKEY()
{
	// since we can't guarantee that the constants for these DIKs won't
	// change, initialize the remap table here instead of statically

	// any key mapped to zero means unmapped and should be ignored
	memset(inw_DIKToINKEY, 0, INW_NUMDIKEYS*sizeof(dword));

	inw_DIKToINKEY[DIK_ESCAPE] = INKEY_ESCAPE;
	inw_DIKToINKEY[DIK_1] = '1';
	inw_DIKToINKEY[DIK_2] = '2';
	inw_DIKToINKEY[DIK_3] = '3';
	inw_DIKToINKEY[DIK_4] = '4';
	inw_DIKToINKEY[DIK_5] = '5';
	inw_DIKToINKEY[DIK_6] = '6';
	inw_DIKToINKEY[DIK_7] = '7';
	inw_DIKToINKEY[DIK_8] = '8';
	inw_DIKToINKEY[DIK_9] = '9';
	inw_DIKToINKEY[DIK_0] = '0';
	inw_DIKToINKEY[DIK_MINUS] = '-';
	inw_DIKToINKEY[DIK_EQUALS] = '=';
	inw_DIKToINKEY[DIK_BACK] = INKEY_BACKSPACE;
	inw_DIKToINKEY[DIK_TAB] = INKEY_TAB;
	inw_DIKToINKEY[DIK_Q] = 'q';
	inw_DIKToINKEY[DIK_W] = 'w';
	inw_DIKToINKEY[DIK_E] = 'e';
	inw_DIKToINKEY[DIK_R] = 'r';
	inw_DIKToINKEY[DIK_T] = 't';
	inw_DIKToINKEY[DIK_Y] = 'y';
	inw_DIKToINKEY[DIK_U] = 'u';
	inw_DIKToINKEY[DIK_I] = 'i';
	inw_DIKToINKEY[DIK_O] = 'o';
	inw_DIKToINKEY[DIK_P] = 'p';
	inw_DIKToINKEY[DIK_LBRACKET] = '[';
	inw_DIKToINKEY[DIK_RBRACKET] = ']';
	inw_DIKToINKEY[DIK_RETURN] = INKEY_ENTER;
	inw_DIKToINKEY[DIK_LCONTROL] = INKEY_LEFTCTRL;
	inw_DIKToINKEY[DIK_A] = 'a';
	inw_DIKToINKEY[DIK_S] = 's';
	inw_DIKToINKEY[DIK_D] = 'd';
	inw_DIKToINKEY[DIK_F] = 'f';
	inw_DIKToINKEY[DIK_G] = 'g';
	inw_DIKToINKEY[DIK_H] = 'h';
	inw_DIKToINKEY[DIK_J] = 'j';
	inw_DIKToINKEY[DIK_K] = 'k';
	inw_DIKToINKEY[DIK_L] = 'l';
	inw_DIKToINKEY[DIK_SEMICOLON] = ';';
	inw_DIKToINKEY[DIK_APOSTROPHE] = '\'';
	inw_DIKToINKEY[DIK_GRAVE] = '`';
	inw_DIKToINKEY[DIK_LSHIFT] = INKEY_LEFTSHIFT;
	inw_DIKToINKEY[DIK_BACKSLASH] = '\\';
	inw_DIKToINKEY[DIK_Z] = 'z';
	inw_DIKToINKEY[DIK_X] = 'x';
	inw_DIKToINKEY[DIK_C] = 'c';
	inw_DIKToINKEY[DIK_V] = 'v';
	inw_DIKToINKEY[DIK_B] = 'b';
	inw_DIKToINKEY[DIK_N] = 'n';
	inw_DIKToINKEY[DIK_M] = 'm';
	inw_DIKToINKEY[DIK_COMMA] = ',';
	inw_DIKToINKEY[DIK_PERIOD] = '.';
	inw_DIKToINKEY[DIK_SLASH] = '/';
	inw_DIKToINKEY[DIK_RSHIFT] = INKEY_RIGHTSHIFT;
	inw_DIKToINKEY[DIK_MULTIPLY] = INKEY_NUMSTAR;
	inw_DIKToINKEY[DIK_LMENU] = INKEY_LEFTALT;
	inw_DIKToINKEY[DIK_SPACE] = INKEY_SPACE;
	inw_DIKToINKEY[DIK_CAPITAL] = INKEY_CAPSLOCK;
	inw_DIKToINKEY[DIK_F1] = INKEY_F1;
	inw_DIKToINKEY[DIK_F2] = INKEY_F2;
	inw_DIKToINKEY[DIK_F3] = INKEY_F3;
	inw_DIKToINKEY[DIK_F4] = INKEY_F4;
	inw_DIKToINKEY[DIK_F5] = INKEY_F5;
	inw_DIKToINKEY[DIK_F6] = INKEY_F6;
	inw_DIKToINKEY[DIK_F7] = INKEY_F7;
	inw_DIKToINKEY[DIK_F8] = INKEY_F8;
	inw_DIKToINKEY[DIK_F9] = INKEY_F9;
	inw_DIKToINKEY[DIK_F10] = INKEY_F10;
	inw_DIKToINKEY[DIK_NUMLOCK] = INKEY_NUMLOCK;
	inw_DIKToINKEY[DIK_SCROLL] = INKEY_SCROLLLOCK;
	inw_DIKToINKEY[DIK_NUMPAD7] = INKEY_NUM7;
	inw_DIKToINKEY[DIK_NUMPAD8] = INKEY_NUM8;
	inw_DIKToINKEY[DIK_NUMPAD9] = INKEY_NUM9;
	inw_DIKToINKEY[DIK_SUBTRACT] = INKEY_NUMMINUS;
	inw_DIKToINKEY[DIK_NUMPAD4] = INKEY_NUM4;
	inw_DIKToINKEY[DIK_NUMPAD5] = INKEY_NUM5;
	inw_DIKToINKEY[DIK_NUMPAD6] = INKEY_NUM6;
	inw_DIKToINKEY[DIK_ADD] = INKEY_NUMPLUS;
	inw_DIKToINKEY[DIK_NUMPAD1] = INKEY_NUM1;
	inw_DIKToINKEY[DIK_NUMPAD2] = INKEY_NUM2;
	inw_DIKToINKEY[DIK_NUMPAD3] = INKEY_NUM3;
	inw_DIKToINKEY[DIK_NUMPAD0] = INKEY_NUM0;
	inw_DIKToINKEY[DIK_DECIMAL] = INKEY_NUMPERIOD;
	inw_DIKToINKEY[DIK_F11] = INKEY_F11;
	inw_DIKToINKEY[DIK_F12] = INKEY_F12;
	inw_DIKToINKEY[DIK_NUMPADENTER] = INKEY_NUMENTER;
	inw_DIKToINKEY[DIK_RCONTROL] = INKEY_RIGHTCTRL;
	inw_DIKToINKEY[DIK_DIVIDE] = INKEY_NUMSLASH;
	inw_DIKToINKEY[DIK_SYSRQ] = INKEY_PRINTSCRN;
	inw_DIKToINKEY[DIK_RMENU] = INKEY_RIGHTALT;
	inw_DIKToINKEY[DIK_HOME] = INKEY_HOME;
	inw_DIKToINKEY[DIK_UP] = INKEY_UPARROW;
	inw_DIKToINKEY[DIK_PRIOR] = INKEY_PGUP;
	inw_DIKToINKEY[DIK_LEFT] = INKEY_LEFTARROW;
	inw_DIKToINKEY[DIK_RIGHT] = INKEY_RIGHTARROW;
	inw_DIKToINKEY[DIK_END] = INKEY_END;
	inw_DIKToINKEY[DIK_DOWN] = INKEY_DOWNARROW;
	inw_DIKToINKEY[DIK_NEXT] = INKEY_PGDN;
	inw_DIKToINKEY[DIK_INSERT] = INKEY_INS;
	inw_DIKToINKEY[DIK_DELETE] = INKEY_DEL;
	inw_DIKToINKEY[DIK_LWIN] = 0;
	inw_DIKToINKEY[DIK_RWIN] = 0;
	inw_DIKToINKEY[DIK_APPS] = 0;
}

static void AcquireMouse()
{
	if (lpMouse)
	{
		HRESULT err;
		int i;

		mouseAcquired = 1;
		if ((err = lpMouse->Acquire()) == DI_OK)
			return;
		// didn't get it back, try a few more times
		for (i=0;i<50;i++)
		{
			if ((err = lpMouse->Acquire()) == DI_OK)
				return;
		}
		// still don't have it, something's up
		mouseAcquired = 0;
	}
}

static void UnacquireMouse()
{
	mouseAcquired = 0;
	if (lpMouse)
		lpMouse->Unacquire();
}

static void AcquireKeyboard()
{
	if (lpKeyboard)
	{
		HRESULT err;
		int i;

		keyboardAcquired = 1;
		if ((err = lpKeyboard->Acquire()) == DI_OK)
			return;
		// didn't get it back, try a few more times
		for (i=0;i<50;i++)
		{
			if ((err = lpKeyboard->Acquire()) == DI_OK)
				return;
		}
		// still don't have it, something's up
		keyboardAcquired = 0;
	}	
		
	FlushKeyboardData();
}

static void UnacquireKeyboard()
{
	FlushKeyboardData();

	keyboardAcquired = 0;
	if (lpKeyboard)
		lpKeyboard->Unacquire();
}

static void CheckKeyEvents(dword key, in_event_t *event)
{
	if (keyStates[key])
	{
		if (!oldKeyStates[key])
		{
			// press event (key was up before but down now)
			event->eventType = INEV_PRESS;
			event->mouseDeltaX = 0;
			event->mouseDeltaY = 0;
			event->pressTimeDelta = 0.0;
            event->lastPressEventDelta = sys_curTime - keyPressTimes[key];
			event->eventIndex = inw_eventIndex++;
			if (inw_eventHandler)
				inw_eventHandler(event);
			keyMouseDragPositions[key][0] = mouseFreeX;
			keyMouseDragPositions[key][1] = mouseFreeY;
			keyPressTimes[key] = sys_curTime;
		}
		else
		{
			// drag event (key is still down)
			event->eventType = INEV_DRAG;
			event->mouseDeltaX = mouseFreeX - keyMouseDragPositions[key][0];
			event->mouseDeltaY = mouseFreeY - keyMouseDragPositions[key][1];
			event->pressTimeDelta = sys_curTime - keyPressTimes[key];
            event->lastPressEventDelta = event->pressTimeDelta;
			event->eventIndex = inw_eventIndex++;
			if (inw_eventHandler)
				inw_eventHandler(event);
			keyMouseDragPositions[key][0] = mouseFreeX;
			keyMouseDragPositions[key][1] = mouseFreeY;
		}
	}
	else if (oldKeyStates[key])
	{
		// release event (key was down before but up now)
		event->eventType = INEV_RELEASE;
		event->mouseDeltaX = mouseFreeX - keyMouseDragPositions[key][0];
		event->mouseDeltaY = mouseFreeY - keyMouseDragPositions[key][1];
        event->pressTimeDelta = sys_curTime - keyPressTimes[key];
        event->lastPressEventDelta = event->pressTimeDelta;
		event->eventIndex = inw_eventIndex++;
		if (inw_eventHandler)
			inw_eventHandler(event);
	}
}

//============================================================================
//    INTERFACE FUNCTIONS
//============================================================================

// initialization
void INW_Init()
{	
	GUID guid;
	DIPROPDWORD prop = { { sizeof(DIPROPDWORD), sizeof(DIPROPHEADER), 0,
		DIPH_DEVICE, }, INW_INPUTBUFFERSIZE };

	// create directinput object
	if (DirectInputCreate(sysw_hInst, DIRECTINPUT_VERSION, &lpDI, NULL) != DI_OK)
		SYS_Errorf("INW_Init failure: DirectInputCreate");

	// init keyboard
	guid = GUID_SysKeyboard;
	if (lpDI->CreateDevice(guid, &lpKeyboard, NULL) != DI_OK)
		SYS_Errorf("INW_Init failure: CreateDevice (keyboard)");

	if (lpKeyboard->SetDataFormat(&c_dfDIKeyboard) != DI_OK)
		SYS_Errorf("INW_Init failure: SetDataFormat (keyboard)");

	// set coop level.  note: the keyboard can't be exclusive
	if (lpKeyboard->SetCooperativeLevel(sysw_hwnd,DISCL_NONEXCLUSIVE|DISCL_FOREGROUND) != DI_OK)
		SYS_Errorf("INW_Init failure: SetCooperativeLevel (keyboard)");

	// set the buffer size to our input buffer maximum
	if (lpKeyboard->SetProperty(DIPROP_BUFFERSIZE, &prop.diph) != DI_OK)
		SYS_Errorf("INW_Init failure: SetProperty (keyboard)");

	// get it
	AcquireKeyboard();

	// set up DIK_ remap
	InitDIKToINKEY();

	// init mouse
	guid = GUID_SysMouse;
	if (lpDI->CreateDevice(guid, &lpMouse, NULL) != DI_OK)
		SYS_Errorf("INW_Init failure: CreateDevice (mouse)");

	if (lpMouse->SetDataFormat(&c_dfDIMouse) != DI_OK)
		SYS_Errorf("INW_Init failure: SetDataFormat (mouse)");

	// set coop level
	if (SetMouseCooperative(DISCL_FOREGROUND) != DI_OK)
		SYS_Errorf("INW_Init failure: SetMouseCooperative");

	// get it
	AcquireMouse();
}

// shutdown
void INW_Shutdown()
{
	// make sure we're initialized
	if (!lpDI)
		return;
	
	// shutdown keyboard
	if (lpKeyboard)
	{
		UnacquireKeyboard();
		lpKeyboard->Release();
		lpKeyboard = NULL;
	}
	
	// shutdown mouse
	if (lpMouse)
	{
		UnacquireMouse();
		lpMouse->Release();
		lpMouse = NULL;
	}
	
	// kill directinput
	lpDI->Release();
	lpDI = NULL;
}

// frame processing - the big function
void INW_Frame()
{
	DIDEVICEOBJECTDATA keys[INW_INPUTBUFFERSIZE];
	DIMOUSESTATE mstate;
	dword key;
	dword i, num;
	sdword oldx, oldy;
	HRESULT err;
	in_event_t event;
	bbool ready;
	static dword keyForMouseButton[3] = { INKEY_MOUSELEFT, INKEY_MOUSERIGHT, INKEY_MOUSEMIDDLE };
	static bbool keyEventSent[INKEY_NUMKEYS];

	//---------------------
	// mouse
	//---------------------	

	ready = 1;
	err = lpMouse->GetDeviceState(sizeof(DIMOUSESTATE), (LPVOID) &mstate);
	if (err != DI_OK)
	{
		if (err==DIERR_NOTACQUIRED)
		{
			// if it's not acquired, get it back and try again
			ready = 0;
			AcquireMouse();
			if ((err = lpMouse->GetDeviceState(sizeof(DIMOUSESTATE), (LPVOID) &mstate)) == DI_OK)
				ready = 1;
			return;
		}
		else if (err==DIERR_INPUTLOST)
		{
			// if the input is gone, all we can do is reacquire
			ready = 0;
			AcquireMouse();
		}
	}
	
	// handle events
	if (ready)
	{
		// get new mouse position status
		oldx = mouseFreeX;
		oldy = mouseFreeY;
		mouseFreeX += mstate.lX * mouseSensitivity;
		mouseFreeY += mstate.lY * mouseSensitivity;
        mouseX += mstate.lX * mouseSensitivity;
        mouseY += mstate.lY * mouseSensitivity;

		// clamp non-free mouse values to limits
		if (mouseX >= mouseMaxX)
			mouseX = mouseMaxX-1;
		if (mouseY >= mouseMaxY)
			mouseY = mouseMaxY-1;
		if (mouseX <= mouseMinX)
			mouseX = mouseMinX+1;
		if (mouseY <= mouseMinY)
			mouseY = mouseMinY+1;

		// rig up event data that won't change in the loop
		event.mouseX = mouseX;
		event.mouseY = mouseY;
		event.mouseFreeX = mouseFreeX;
		event.mouseFreeY = mouseFreeY;
		event.mouseDeltaX = event.mouseDeltaY = 0;
		event.frameTimeStamp = sys_curTime;
		event.pressTimeDelta = 0.0;
		event.key = 0;
		event.flags = 0; // note: key flags will be off by a frame here since the mouse goes first
		if (keyStates[INKEY_LEFTSHIFT] || keyStates[INKEY_RIGHTSHIFT])
			event.flags |= INEVF_SHIFT;
		if (keyStates[INKEY_LEFTCTRL] || keyStates[INKEY_RIGHTCTRL])
			event.flags |= INEVF_CTRL;
		if (keyStates[INKEY_LEFTALT] || keyStates[INKEY_RIGHTALT])
			event.flags |= INEVF_ALT;
		
		// call a MOUSEMOVE event first if one occured
		if ((mouseFreeX != oldx) || (mouseFreeY != oldy))
		{
			event.mouseDeltaX = mouseFreeX - oldx;
			event.mouseDeltaY = mouseFreeY - oldy;
			event.eventType = INEV_MOUSEMOVE;
			event.eventIndex = inw_eventIndex++;
			if (inw_eventHandler)
				inw_eventHandler(&event);
		}
		
		// check the mouse buttons
		for(i=0; i<3; i++)
		{
			key = keyForMouseButton[i];
			event.key = key;

			// set current key state
			keyStates[key] = mstate.rgbButtons[i] & (byte)0x80;

			CheckKeyEvents(key, &event);

			// save current key state to old
			oldKeyStates[key] = keyStates[key];
		}
	}

	//---------------------
	// keyboard
	//---------------------
	
	ready = 1;
	num=INW_INPUTBUFFERSIZE;
	err=lpKeyboard->GetDeviceData(sizeof(DIDEVICEOBJECTDATA),keys,&num,0);
	if (err != DI_OK)
	{
		if (err==DIERR_NOTACQUIRED)
		{
			// if it's not acquired, get it back and try again
			ready = 0;
			AcquireKeyboard();
			num=INW_INPUTBUFFERSIZE;
			if ((err=lpKeyboard->GetDeviceData(sizeof(DIDEVICEOBJECTDATA),keys,&num,0)) == DI_OK)
				ready = 1;
		}
		else if (err==DIERR_INPUTLOST)
		{
			// if the input is gone, all we can do is reacquire
			ready = 0;
			AcquireKeyboard();
		}
	}
	
	// handle events
	if (ready)
	{
		event.mouseX = mouseX;
		event.mouseY = mouseY;
		event.mouseFreeX = mouseFreeX;
		event.mouseFreeY = mouseFreeY;
		event.frameTimeStamp = sys_curTime;

		// run through input buffer to account for changed states
		for (i=0;i<num;i++)
		{
			key = inw_DIKToINKEY[keys[i].dwOfs]; // convert to our key table
			if (!key)
				continue; // not a mapped key, ignore it
			
			// set current key state
			keyStates[key] = (byte)keys[i].dwData & 0x80L;

			if (keyStates[key] != oldKeyStates[key])
			{
				// press or release event, need to handle it now
				event.key = key;
				event.flags = 0;
				if (keyStates[INKEY_LEFTSHIFT] || keyStates[INKEY_RIGHTSHIFT])
					event.flags |= INEVF_SHIFT;
				if (keyStates[INKEY_LEFTCTRL] || keyStates[INKEY_RIGHTCTRL])
					event.flags |= INEVF_CTRL;
				if (keyStates[INKEY_LEFTALT] || keyStates[INKEY_RIGHTALT])
					event.flags |= INEVF_ALT;

				CheckKeyEvents(key, &event);

				// save current key state to old
				oldKeyStates[key] = keyStates[key];

				// make sure we don't call a drag event in this frame
				keyEventSent[key] = 1;
			}
		}

		// handle key injection buffer in the same way
		for (i=0;i<keyInjectionBufferIndex;i++)
		{
			key = keyInjectionBuffer[i][0];

			// set current key state
			keyStates[key] = keyInjectionBuffer[i][1];

			if (keyStates[key] != oldKeyStates[key])
			{
				// press or release event, need to handle it now
				event.key = key;
				event.flags = 0;
				if (keyStates[INKEY_LEFTSHIFT] || keyStates[INKEY_RIGHTSHIFT])
					event.flags |= INEVF_SHIFT;
				if (keyStates[INKEY_LEFTCTRL] || keyStates[INKEY_RIGHTCTRL])
					event.flags |= INEVF_CTRL;
				if (keyStates[INKEY_LEFTALT] || keyStates[INKEY_RIGHTALT])
					event.flags |= INEVF_ALT;

				CheckKeyEvents(key, &event);

				// save current key state to old
				oldKeyStates[key] = keyStates[key];

				// make sure we don't call a drag event in this frame
				keyEventSent[key] = 1;
			}
		}
		keyInjectionBufferIndex = 0;

		// send drag events for all unhandled keys
		event.flags = 0;
		if (keyStates[INKEY_LEFTSHIFT] || keyStates[INKEY_RIGHTSHIFT])
			event.flags |= INEVF_SHIFT;
		if (keyStates[INKEY_LEFTCTRL] || keyStates[INKEY_RIGHTCTRL])
			event.flags |= INEVF_CTRL;
		if (keyStates[INKEY_LEFTALT] || keyStates[INKEY_RIGHTALT])
			event.flags |= INEVF_ALT;
		for (i=0;i<256;i++)
		{
			if (keyEventSent[i] || !keyStates[i])
				continue; // already ran events on this key, or no drag to deal with
			event.key = i;
			
			CheckKeyEvents(i, &event);

			// save current key state to old
			oldKeyStates[i] = keyStates[i];
		}

		// clear the eventsent buffer
		memset(keyEventSent, 0, INKEY_NUMKEYS);
	}
}

// called by sys_win when focus is restored
void INW_SetFocus()
{
	AcquireMouse();
	AcquireKeyboard();
}

// called by sys_win when focus is lost
void INW_KillFocus()
{
	UnacquireMouse();
	UnacquireKeyboard();
}

// sets the event callback which INW_Frame uses
void INW_SetEventHandler(void (*handler)(in_event_t *))
{
	inw_eventHandler = handler;
}

// force a key event into the event stream which is processed at the end
// of the next INW_Frame call.
void INW_InjectKeyEvent(dword key, bbool state)
{
	if (keyInjectionBufferIndex >= INW_MAXINJECT)
		return; // too many injected keys
	keyInjectionBuffer[keyInjectionBufferIndex][0] = key;
	keyInjectionBuffer[keyInjectionBufferIndex][1] = state;
	keyInjectionBufferIndex++;
}

// change the exclusive mode on the mouse
void INW_SetMouseExclusive(bbool exclusive)
{
	bbool wasAcquired = mouseAcquired;

	mouseExclusive = exclusive;	
	if (wasAcquired)
		UnacquireMouse();
	if (SetMouseCooperative(DISCL_FOREGROUND) != DI_OK)
		SYS_Errorf("INW_SetMouseExclusive failure");
	if (wasAcquired)
		AcquireMouse();
}

// set the limit box for the mouseX/mouseY values
void INW_SetMouseLimits(sdword x1, sdword y1, sdword x2, sdword y2)
{
	mouseMinX = x1;
	mouseMinY = y1;
	mouseMaxX = x2;
	mouseMaxY = y2;
}

// set the mouse sensitivity multiplier
void INW_SetMouseSensitivity(float factor)
{
	mouseSensitivity = factor;
}

float INW_GetMouseSensitivity()
{
	return(mouseSensitivity);
}

//============================================================================
//    INTERFACE CLASS BODIES
//============================================================================

//****************************************************************************
//**
//**    END MODULE IN_WIN.CPP
//**
//****************************************************************************

