//****************************************************************************
//**
//**    MEM_MAIN.CPP
//**    Memory Management
//**
//**    Copyright (c) 1998 Chris Hargrove
//**    Notice: Usage of any code in this file is subject to the rules
//**            described in the LICENSE.TXT file included in this directory.
//**            Reading, compiling, or otherwise using this code constitutes
//**            automatic acceptance of the rules in said text file.
//**
//****************************************************************************
//============================================================================
//    IMPLEMENTATION HEADERS
//============================================================================
#define SYS_DEFS_INCLUDE_WIN32  // for SYSW_Mem* functions
#include "sys_defs.h"
#include "mem_main.h"
//============================================================================
//    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
//============================================================================
#define MEM_MAXZONES            512 // maximum number of active zones
#define MEM_DEFAULTZONEBYTES    (4*1024*1024) // size in bytes of default zone

// zone flags
enum
{
    MEMZONEF_INUSE = 0x00000001 // zone is currently active
};

//============================================================================
//    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
//============================================================================

// zone structure
typedef struct
{
    dword flags;
    byte* basePtr; // base address of zone
    dword curSize; // current size of zone
    dword maxSize; // size limit of zone
} mem_zone_t;

//============================================================================
//    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE DATA
//============================================================================
mem_zone_t mem_Zones[MEM_MAXZONES]; // zone pool
dword mem_DefaultZone; // zone used when zone of zero is passed during alloc

//============================================================================
//    INTERFACE DATA
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
//============================================================================
static mem_zone_t* AllocZone();
static void FreeZone(mem_zone_t *zone);
static mem_zone_t* ZoneForId(dword zoneid);
static dword IdForZone(mem_zone_t* zone);

//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTIONS
//============================================================================

// allocate a zone structure from zone pool
static mem_zone_t* AllocZone()
{
    dword i;
    for (i=0;i<MEM_MAXZONES;i++)
    {
        if (!(mem_Zones[i].flags & MEMZONEF_INUSE))
        {
            mem_Zones[i].flags = MEMZONEF_INUSE;
            mem_Zones[i].basePtr = NULL;
            mem_Zones[i].curSize = 0;
            mem_Zones[i].maxSize = 0;
            return(&mem_Zones[i]);
        }
    }
    return(NULL);
}

// free a zone structure back to pool
static void FreeZone(mem_zone_t *zone)
{
    if (!zone)
        return;
    zone->flags &= ~MEMZONEF_INUSE;
}

// zone<->zoneid exchange functions
static mem_zone_t* ZoneForId(dword zoneid)
{
    if ((!zoneid) || (zoneid > MEM_MAXZONES))
        return(NULL);
    if (!(mem_Zones[zoneid-1].flags & MEMZONEF_INUSE))
        return(NULL);
    return(&mem_Zones[zoneid-1]);
}

static dword IdForZone(mem_zone_t* zone)
{
    return((dword)(zone - mem_Zones)+1);
}

//============================================================================
//    INTERFACE FUNCTIONS
//============================================================================

// init/shutdown

void MEM_Init()
{
    dword defAlloc=0;
    byte parm;

    if (parm = M_CmdLineOption("defzonesize", 1))
        defAlloc = atoi(SYS_Argv(parm+1))*1024*1024;
    if (!defAlloc)
        defAlloc=MEM_DEFAULTZONEBYTES;
    if (!(mem_DefaultZone=MEM_ZoneCreate(defAlloc)))
        SYS_Errorf("MEM_Init: Unable to create default zone");
}

void MEM_Shutdown()
{
    dword i;
    for (i=1;i<=MEM_MAXZONES;i++)
    {
        if (MEM_ZoneIsValid(i))
            MEM_ZoneDestroy(i);
    }
}

// zone functions

// create a new zone with a given size limit
dword MEM_ZoneCreate(dword size)
{
    mem_zone_t* zone;

    if (!(zone=AllocZone()))
        SYS_Errorf("MEM_ZoneStart: AllocZone failure");
    zone->curSize = 0;
    zone->maxSize = size;
    zone->basePtr = (byte*)SYSW_MemReserve(size);
    if (!zone->basePtr)
        SYS_Errorf("MEM_ZoneStart: SYSW_MemReserve failed on %d bytes", size);
    return(IdForZone(zone));
}

// allocate a block of memory from the given zone
void* MEM_ZoneAlloc(dword size, dword zoneid)
{
    void *res;
    mem_zone_t* zone;

    if (!zoneid)
        zoneid = mem_DefaultZone; // allocate from default zone
    if (!(zone=ZoneForId(zoneid)))
        SYS_Errorf("MEM_ZoneAlloc: ZoneForId failure (zone %d)", zoneid);
    size = (size+0x1F) & ~0x1F; // cacheline
    zone->curSize += size;
    if (zone->curSize > zone->maxSize)
        SYS_Errorf("MEM_ZoneAlloc overflow (%d > %d)", zone->curSize, zone->maxSize);
    if (!(res = (byte*)SYSW_MemCommit(zone->basePtr, zone->curSize)))
        SYS_Errorf("MEM_ZoneAlloc: SYS_MemCommit failed on %d bytes", zone->curSize);
    return (void *)(zone->basePtr + zone->curSize - size);
}

// free a zone and all its allocated memory
void MEM_ZoneDestroy(dword zoneid)
{
    mem_zone_t* zone;

    if (!(zone=ZoneForId(zoneid)))
        SYS_Errorf("MEM_ZoneDestroy: ZoneForId failure (zone %d)", zoneid);
    if (zone->basePtr)
        SYSW_MemFree(zone->basePtr);
    FreeZone(zone);
}

// set a zone back to its base address
void MEM_ZoneRestart(dword zoneid)
{
    mem_zone_t* zone;
    
    if (!(zone=ZoneForId(zoneid)))
        SYS_Errorf("MEM_ZoneRestart: ZoneForId failure (zone %d)", zoneid);
    zone->curSize = 0;
}

// make sure a zoneid is valid
bbool MEM_ZoneIsValid(dword zoneid)
{
    return(ZoneForId(zoneid)!=NULL);
}

// zone info functions
void* MEM_ZoneGetBase(dword zoneid)
{
    mem_zone_t* zone;

    if (!(zone=ZoneForId(zoneid)))
        SYS_Errorf("MEM_ZoneGetBase: ZoneForId failure (zone %d)", zoneid);
    return(zone->basePtr);
}

dword MEM_ZoneGetCurrent(dword zoneid)
{
    mem_zone_t* zone;

    if (!(zone=ZoneForId(zoneid)))
        SYS_Errorf("MEM_ZoneGetCurrent: ZoneForId failure (zone %d)", zoneid);
    return(zone->curSize);
}

dword MEM_ZoneGetLimit(dword zoneid)
{
    mem_zone_t* zone;

    if (!(zone=ZoneForId(zoneid)))
        SYS_Errorf("MEM_ZoneGetLimit: ZoneForId failure (zone %d)", zoneid);
    return(zone->maxSize);
}

//============================================================================
//    INTERFACE CLASS BODIES
//============================================================================

//****************************************************************************
//**
//**    END MODULE MEM_MAIN.CPP
//**
//****************************************************************************

