//****************************************************************************
//**
//**    RES_MAN.CPP
//**    Resource Manager
//**
//**    Copyright (c) 1998 Chris Hargrove
//**    Notice: Usage of any code in this file is subject to the rules
//**            described in the LICENSE.TXT file included in this directory.
//**            Reading, compiling, or otherwise using this code constitutes
//**            automatic acceptance of the rules in said text file.
//**
//****************************************************************************
//============================================================================
//    IMPLEMENTATION HEADERS
//============================================================================
#include "sys_defs.h"
#include "res_man.h"

//============================================================================
//    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
//============================================================================
#define RES_MAX_ENTRYBANKS           1024
#define RES_MAX_ENTRYBANKENTRIES     256

#define RES_TEMPZONEBYTES     (1*1024*1024) // size in bytes of temp zone
#define RES_LEVELZONEBYTES    (8*1024*1024) // size in bytes of level zone
#define RES_GAMEZONEBYTES     (3*1024*1024) // size in bytes of game zone

//============================================================================
//    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
//============================================================================
// resource entry
typedef struct
{
    res_id_t id; // resource id, low word is entry, high word is bank
    char fileName[SYS_MAX_PATH]; // string file name
    res_cacheType_t cacheType; // cache type
    res_loaderFunc_t loadFunc; // loader function called on cachein
    res_freeFunc_t freeFunc; // freeing function called on cacheout
    void* dataPtr; // data pointer, null when cached out
    dword dataSize; // data size, 0 when cached out
    dword fileNameTag; // tag value as a lame hash for filename
                       // low word is sum, high word is length
} res_entry_t;

// resource entry bank, contains a set of entries when needed
typedef struct
{
    dword entryCount;
    res_entry_t *entries;
} res_entryBank_t;

//============================================================================
//    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE DATA
//============================================================================
static int res_mallocCache = -1; // RES_Malloc cache type, set during loads only
static dword res_cacheZones[RES_CACHE_NUMTYPES]; // zones for each cache type
static dword res_banksUsed = 0; // number of full banks
static res_entryBank_t res_Banks[RES_MAX_ENTRYBANKS]; // bank pool

//============================================================================
//    INTERFACE DATA
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
//============================================================================
static res_entry_t* EntryForId(res_id_t id);
static res_entry_t* FindEntry(char* fileName, dword fileNameTag);
static res_entry_t* AllocEntry();
static void CacheInEntry(res_entry_t* entry);
static void CacheOutEntry(res_entry_t* entry);
static void* DefaultLoadFunc(char* fileName);

//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTIONS
//============================================================================

// get the entry for a resource id
static res_entry_t* EntryForId(res_id_t id)
{
    res_entryBank_t* bank;
    dword bankIndex = (id>>16)&0xFFFF;
    dword entryIndex = id&0xFFFF;
    
    if (bankIndex > res_banksUsed)
        SYS_Errorf("EntryForId: Excessive Bank Index %d", bankIndex);
    bank = &res_Banks[bankIndex];
    if (entryIndex >= bank->entryCount)
        SYS_Errorf("EntryForId: Excessive Entry Index %d in Bank %d",
            entryIndex, bankIndex);
    return(&bank->entries[entryIndex]);
}

// find an existing entry matching a given name and tag
static res_entry_t* FindEntry(char* fileName, dword fileNameTag)
{
    res_entryBank_t* bank;
    res_entry_t* entry;
    
    for (dword i=0;i<=res_banksUsed;i++)
    {
        bank = &res_Banks[i];
        for (dword k=0;k<bank->entryCount;k++)
        {        
            entry = &bank->entries[k];
            if (fileNameTag != entry->fileNameTag)
                continue;
            if (strcmp(fileName, entry->fileName))
                continue;
            return(&bank->entries[k]);
        }
    }
    return(NULL);
}

// allocate a new entry
static res_entry_t* AllocEntry()
{
    res_entryBank_t* bank = &res_Banks[res_banksUsed];
    if (!bank->entries)
        bank->entries = SYS_MALLOC(res_entry_t, RES_MAX_ENTRYBANKENTRIES);
    res_entry_t* entry = &bank->entries[bank->entryCount];
    entry->id = (res_banksUsed<<16)+bank->entryCount;
    bank->entryCount++;
    if (bank->entryCount >= RES_MAX_ENTRYBANKENTRIES)
        res_banksUsed++;
    return(entry);
}

// make sure an entry's data is cached in
static void CacheInEntry(res_entry_t* entry)
{
    if (entry->dataPtr)
        return; // already cached in
    entry->dataSize = 0;
    dword mallocStart = MEM_ZoneGetCurrent(res_cacheZones[entry->cacheType]);
    res_mallocCache = entry->cacheType;
    if (!(entry->dataPtr = entry->loadFunc(entry->fileName)))
    {
        // loader failed on regular filename, try default
        char buf[SYS_MAX_PATH];
        strcpy(buf, "DEFAULT.");
        strcat(buf, M_GetFileExtension(entry->fileName));
        if (!(entry->dataPtr = entry->loadFunc(buf)))
            SYS_Errorf("CacheInEntry: Could not load %s nor %s", entry->fileName, buf);
    }
    res_mallocCache = -1;
    entry->dataSize = MEM_ZoneGetCurrent(res_cacheZones[entry->cacheType])
        - mallocStart;
}

// cache entry data out (doesn't need to remove memory, just calls free function)
static void CacheOutEntry(res_entry_t* entry)
{
    if (!entry->dataPtr)
        return; // already cached out
    if (entry->freeFunc)
        entry->freeFunc(entry->id);
    entry->dataPtr = NULL;
    entry->dataSize = 0;
}

// default loader function, loads raw file stream
static void* DefaultLoadFunc(char* fileName)
{
    dword fp = XF_Open(fileName);
    if (!fp)
        return(NULL);
    dword size = XF_Size(fp);
    byte* buf = (byte*)RES_Malloc(size);
    XF_Read(buf, size, fp);
    XF_Close(fp);
    return(buf);
}

//============================================================================
//    INTERFACE FUNCTIONS
//============================================================================

// init/shutdown/frame

void RES_Init()
{
    dword tempReserve=0, levelReserve=0, gameReserve=0;
    byte parm;

    if (parm = M_CmdLineOption("rescache", 3))
    {
        tempReserve = atoi(SYS_Argv(parm+1))*1024*1024;
        levelReserve = atoi(SYS_Argv(parm+2))*1024*1024;
        gameReserve = atoi(SYS_Argv(parm+3))*1024*1024;
    }
    if (!tempReserve)
        tempReserve=RES_TEMPZONEBYTES;
    if (!levelReserve)
        levelReserve=RES_LEVELZONEBYTES;
    if (!gameReserve)
        gameReserve=RES_GAMEZONEBYTES;
    
    // create zones
    res_cacheZones[RES_CACHE_TEMP] = MEM_ZoneCreate(tempReserve);
    res_cacheZones[RES_CACHE_LEVEL] = MEM_ZoneCreate(levelReserve);
    res_cacheZones[RES_CACHE_GAME] = MEM_ZoneCreate(gameReserve);
}

void RES_Shutdown()
{
    res_entryBank_t* bank;
    
    // cache out all resources
    for (dword i=0;i<=res_banksUsed;i++)
    {
        bank = &res_Banks[i];
        for (dword k=0;k<bank->entryCount;k++)
        {        
            CacheOutEntry(&bank->entries[k]);
        }
    }
    // empty banks
    for (i=0;i<RES_MAX_ENTRYBANKS;i++)
    {
        if (res_Banks[i].entries)
            SYS_FREE(res_Banks[i].entries);
    }
    // eliminate zones
    MEM_ZoneDestroy(res_cacheZones[RES_CACHE_TEMP]);
    MEM_ZoneDestroy(res_cacheZones[RES_CACHE_LEVEL]);
    MEM_ZoneDestroy(res_cacheZones[RES_CACHE_GAME]);
}

void RES_Frame()
{
    res_entryBank_t* bank;
    res_entry_t* entry;
    
    // cache out all temp resources
    for (dword i=0;i<=res_banksUsed;i++)
    {
        bank = &res_Banks[i];
        for (dword k=0;k<bank->entryCount;k++)
        {        
            entry = &bank->entries[k];
            if (entry->cacheType == RES_CACHE_TEMP)
                CacheOutEntry(entry);
        }
    }
    // restart the temp zone cache
    MEM_ZoneRestart(res_cacheZones[RES_CACHE_TEMP]);
}

// resource registration

res_id_t RES_Register(char* fileName, res_cacheType_t cacheType,
                      res_loaderFunc_t loadFunc,
                      res_freeFunc_t freeFunc)
{
    static char fileNameUpr[1024];
    dword fileNameTag = 0;
    res_entry_t* entry;

    // if loadFunc is null, use the default raw file loader
    if (!loadFunc)
        loadFunc = DefaultLoadFunc;
    
    // all entry filenames are uppercase
    strcpy(fileNameUpr, fileName);
    strupr(fileNameUpr);
    
    // calculate the lame hash tag value
    for (char* ptr=fileNameUpr; *ptr; ptr++)
        fileNameTag += *ptr;
    fileNameTag += strlen(fileNameUpr)<<16;
    
    // see if the entry already exists
    entry = FindEntry(fileNameUpr, fileNameTag);
    if (!entry)
    {
        // entry didn't exist, create it
        entry = AllocEntry();
        strcpy(entry->fileName, fileNameUpr);
        entry->cacheType = cacheType;
        entry->loadFunc = loadFunc;
        entry->freeFunc = freeFunc;
        entry->dataPtr = NULL;
        entry->dataSize = 0;
        entry->fileNameTag = fileNameTag;
        return(entry->id);
    }

    // entry already exists, but update its loaders
    entry->loadFunc = loadFunc;
    entry->freeFunc = freeFunc;

    // see if entry's cache type is being promoted
    if (cacheType > entry->cacheType)
    {
        // promotion, forget about lower-cache resource and raise it
        entry->cacheType = cacheType;
        entry->dataPtr = NULL;
        entry->dataSize = 0;
    }

    return(entry->id);
}

// resource retrieval

void* RES_PtrForId(res_id_t id, dword* size)
{
    res_entry_t* entry = EntryForId(id);
    CacheInEntry(entry);
    if (size)
        *size = entry->dataSize;
    return(entry->dataPtr);
}

char* RES_FileNameForId(res_id_t id)
{
    return(EntryForId(id)->fileName);
}

// cache control
void RES_LevelChange()
{
    res_entryBank_t* bank;
    res_entry_t* entry;
    
    // cache out all level resources
    for (dword i=0;i<=res_banksUsed;i++)
    {
        bank = &res_Banks[i];
        for (dword k=0;k<bank->entryCount;k++)
        {        
            entry = &bank->entries[k];
            if (entry->cacheType == RES_CACHE_LEVEL)
                CacheOutEntry(entry);
        }
    }
    // restart the level zone cache
    MEM_ZoneRestart(res_cacheZones[RES_CACHE_LEVEL]);
}

// loader function memory allocation, only works in loaders
void* RES_Malloc(dword size)
{
    if (res_mallocCache == -1)
        SYS_Errorf("RES_Malloc called outside of loader routine");
    return(MEM_ZoneAlloc(size, res_cacheZones[res_mallocCache]));
}

//============================================================================
//    INTERFACE CLASS BODIES
//============================================================================

//****************************************************************************
//**
//**    END MODULE RES_MAN.CPP
//**
//****************************************************************************

