//****************************************************************************
//**
//**    SYS_MAIN.CPP
//**    System Main
//**
//**    Copyright (c) 1998 Chris Hargrove
//**    Notice: Usage of any code in this file is subject to the rules
//**            described in the LICENSE.TXT file included in this directory.
//**            Reading, compiling, or otherwise using this code constitutes
//**            automatic acceptance of the rules in said text file.
//**
//****************************************************************************
//============================================================================
//    IMPLEMENTATION HEADERS
//============================================================================
#define SYS_DEFS_INCLUDE_WIN32 // for SYSW_Argc, SYSW_Argv,
                               // and SYSW_MessageBox
#include "sys_defs.h"
#include "sys_main.h"

// temporarily include vid_win for demonstration purposes
#include "vid_win.h"

//============================================================================
//    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
//============================================================================
#define SAFEMALLOC_MARKER 0x2b3d1234
#define SAFEMALLOC_USEMARKER // enable malloc/free marker verify

//============================================================================
//    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
//============================================================================
//============================================================================
//    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE DATA
//============================================================================
static float sys_baseTime;
static float sys_lastTime;

//============================================================================
//    INTERFACE DATA
//============================================================================
char sys_programPath[SYS_MAX_PATH];
float sys_curTime, sys_frameTime;

//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
//============================================================================
static void UpdateFrameTimes();
static void ResetFrameTimes();

//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTIONS
//============================================================================
static void UpdateFrameTimes()
{
	sys_curTime = M_TimeFloat() - sys_baseTime;
	sys_frameTime = sys_curTime - sys_lastTime;
	sys_lastTime = sys_curTime;
}

static void ResetFrameTimes()
{
	sys_baseTime = M_TimeFloat();
	sys_lastTime = sys_curTime = sys_frameTime = 0.0;
}

static bbool TestReceiver(in_event_t* event)
{
	if (event->key == INKEY_ESCAPE)
        SYS_Quit();
    return(0);
}

//============================================================================
//    INTERFACE FUNCTIONS
//============================================================================

// The "Big Three" control functions

// initialization
void SYS_Init()
{
    MEM_Init(); // initialize memory system
	XF_Init(); // initialize file system
    RES_Init(); // initialize resource manager

	if (!M_CmdLineOption("noalert", 0)) // example of command line option use
	{
		SYS_Alertf("Hi!  I'm a small test of our initial DirectDraw code.\n\n"
				   "I'll change into 640x480x8 resolution (if possible) and "
				   "display a series of colored lines.\nHit ESC when "
				   "you want to end the program.\n\n"
				   "BTW, you can run this program with the -noalert "
				   "option to skip this message.");
	}
	
	IN_Init(); // initialize user input
    IN_AddReceiver(TestReceiver);

	SYS_InitDebug();

	//-----------------------------------------
	// initialize our directdraw test

	{
		byte pal[768];
		dword i;

		// initialize the directdraw layer
		VIDW_Init();

		// change the video mode
		if (!VIDW_SetMode(640, 480, 8))
			SYS_Errorf("SYS_Main: Could not set video mode");

		// set the palette to a series of gray, red, green, and blue
		memset(pal, 0, 768);
		for (i=0;i<64;i++)
		{
			pal[i*3] = pal[(i+64)*3] = i<<2;
			pal[i*3+1] = pal[(i+128)*3+1] = i<<2;
			pal[i*3+2] = pal[(i+192)*3+2] = i<<2;
		}
		VIDW_SetPaletteRange(pal, 0, 256);
	}

	//-----------------------------------------

	ResetFrameTimes(); // initialize frame times from this point
}

// main loop
void SYS_Frame()
{
	UpdateFrameTimes();
	
	RES_Frame(); // update resources
    IN_Frame(); // process user input

	//-----------------------------------------
	// run the frame for our directdraw test

	{
		static dword iteration = 0;
		dword i;
		byte* buf;
		
		// get the offscreen buffer
		buf = VIDW_GetBuffer();
		// write a series of lines of incremental color
		for (i=0;i<VIDW_GetHeight();i++)
			memset(buf+i*VIDW_GetPitch(), (i+iteration)&255, VIDW_GetWidth());
		// increate the iteration so the lines will move
		iteration++;
		// call the frame blitter
		VIDW_Frame();
	}

	//-----------------------------------------
}

// de-initialization
void SYS_Shutdown()
{	
    static bbool shuttingDown = 0;
    if (shuttingDown)
        return; // recursive shutdown block
    shuttingDown = 1;

	//-----------------------------------------
    // shut down our directdraw test

	VIDW_Shutdown();
	
	//-----------------------------------------

	SYS_ShutdownDebug();

	IN_Shutdown();
    RES_Shutdown();
    XF_Shutdown();
    MEM_Shutdown();
}

// alert notification
void SYS_Alertf(char* text, ...)
{
	static char textBuffer[2048];
	va_list args;
	va_start(args, text);
	vsprintf(textBuffer, text, args);
	va_end(args);
	SYSW_MessageBox(APPTITLE " Notification", MB_OK, textBuffer);
}

// initialize debugging window
void SYS_InitDebug()
{
	SYSW_InitDebug();
}

// shut down debugging window
void SYS_ShutdownDebug()
{
	SYSW_ShutdownDebug();
}

// debugging window output
void SYS_Debugf(char* text, ...)
{
	static char textBuffer[1024];
	va_list args;
	va_start(args, text);
	vsprintf(textBuffer, text, args);
	va_end(args);
	SYSW_DebugString(textBuffer);
}

// exit without error
void SYS_Quit()
{
	SYS_Shutdown();
	exit(0);
}

// exit with fatal error
void SYS_Errorf(char *text, ...)
{
	static char textBuffer[1024];
	va_list args;
	va_start(args, text);
	vsprintf(textBuffer, text, args);
	va_end(args);
	SYSW_MessageBox("Fatal Error", MB_OK, textBuffer);
	SYS_Shutdown();
	exit(1);
}

// command line argument functions, call windows equivalents
byte SYS_Argc()
{
	return(SYSW_Argc());
}

char* SYS_Argv(byte index)
{
	return(SYSW_Argv(index));
}

// safe malloc and free wrappers, with optional marker verify (4 byte cost)
void* SYS_SafeMalloc(int size)
{
#ifdef SAFEMALLOC_USEMARKER	
	size += 4;
#endif
	void *res = malloc(size);
	if (!res)
		SYS_Errorf("SYS_SafeMalloc: Out of memory on %d bytes", size);
#ifdef SAFEMALLOC_USEMARKER	
	*((long *)res) = SAFEMALLOC_MARKER;
	return((long *)res+1);
#else
	return(res);
#endif
}

void SYS_SafeFree(void **ptr)
{
	long *p;
	if (!(*ptr))
		SYS_Errorf("SYS_SafeFree: Attempt to free null pointer");
	p = (long *)*ptr;
#ifdef SAFEMALLOC_USEMARKER	
	p -= 1;
	if (*p != SAFEMALLOC_MARKER)
    {
		if (*p == (SAFEMALLOC_MARKER ^ 0xFFFFFFFF))
            SYS_Errorf("SYS_SafeFree: Attempt to free pointer twice");
        else
            SYS_Errorf("SYS_SafeFree: Attempt to free non-SafeMalloc'ed pointer");
    }
    *p = SAFEMALLOC_MARKER ^ 0xFFFFFFFF;
#endif
	free(p);
	*ptr = NULL;
}

//============================================================================
//    INTERFACE CLASS BODIES
//============================================================================

//****************************************************************************
//**
//**    END MODULE SYS_MAIN.CPP
//**
//****************************************************************************

