//****************************************************************************
//**
//**    VID_WIN.CPP
//**    Windows DirectDraw
//**
//**    Copyright (c) 1998 Chris Hargrove
//**    Notice: Usage of any code in this file is subject to the rules
//**            described in the LICENSE.TXT file included in this directory.
//**            Reading, compiling, or otherwise using this code constitutes
//**            automatic acceptance of the rules in said text file.
//**
//****************************************************************************
//============================================================================
//    IMPLEMENTATION HEADERS
//============================================================================
#define SYS_DEFS_INCLUDE_WIN32
#include "sys_defs.h"
#include "vid_win.h"

#include <ddraw.h>
//============================================================================
//    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
//============================================================================
#define VID_MAXDEVICES  32		// maximum number of devices to use
#define VID_MAXMODES    1024	// maximum number of modes to use

//============================================================================
//    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
//============================================================================

// video device structure, represents a directdraw device
typedef struct
{
    dwbool isNullGuid;
    GUID deviceGuid;
    char name[248];
    char description[256];
    DDCAPS devCaps;
    DDCAPS helCaps;
} vid_device_t;

// video mode structure
typedef struct
{
    vid_device_t *device;
    dword width;
    dword height;
    dword bitDepth;
    dword refresh;
    DDSURFACEDESC surfDesc;
} vid_mode_t;
//============================================================================
//    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE DATA
//============================================================================
// directdraw variables
static LPDIRECTDRAW2 vid_lpDD2 = NULL;
static LPDIRECTDRAWSURFACE vid_lpDDSPrimary = NULL;
static LPDIRECTDRAWSURFACE vid_lpDDSBack = NULL;
static LPDIRECTDRAWSURFACE vid_lpDDSOffscreen = NULL;
static LPDIRECTDRAWPALETTE vid_lpDDPalette = NULL;

// device and mode pools
static dword vid_numDevices = 0;
static vid_device_t vid_Devices[VID_MAXDEVICES];
static dword vid_numModes = 0;
static vid_mode_t vid_Modes[VID_MAXMODES];

// status values
static vid_device_t* vid_currentDevice = NULL;
static vid_mode_t* vid_currentMode = NULL;
static dword vid_ScreenWidth, vid_ScreenHeight, vid_ScreenBitDepth;
static dword vid_ScreenPitch;
static byte* vid_ScreenPtr;
static byte vid_Palette[768];

//============================================================================
//    INTERFACE DATA
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
//============================================================================
static void ModeShutdown();
static void DeviceShutdown();
static void ActivateDevice(vid_device_t* device);
static void ActivateMode(vid_mode_t* mode);
//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTIONS
//============================================================================

// shut down the current video mode
static void ModeShutdown()
{
	// destroy surfaces and palette created at mode activation
	if (vid_lpDDSOffscreen)
	{
		vid_lpDDSOffscreen->Unlock(vid_ScreenPtr);
		vid_lpDDSOffscreen->Release();
		vid_lpDDSOffscreen = NULL;
	}
	if (vid_lpDDSBack)
	{
		vid_lpDDSBack->Release();
		vid_lpDDSBack = NULL;
	}
	if (vid_lpDDSPrimary)
	{
		vid_lpDDSPrimary->Release();
		vid_lpDDSPrimary = NULL;
	}	
	if (vid_lpDDPalette)
	{
		vid_lpDDPalette->Release();
		vid_lpDDPalette = NULL;
	}

	// set the cooperative level back
	if (vid_lpDD2)
		vid_lpDD2->SetCooperativeLevel(sysw_hwnd, DDSCL_NORMAL);

    vid_currentMode = NULL;
}

// shut down the current device
static void DeviceShutdown()
{
	// release the directdraw interface
	if (vid_lpDD2)
	{
		vid_lpDD2->Release();
		vid_lpDD2 = NULL;
	}

    vid_currentDevice = NULL;
}

// activate a directdraw device
static void ActivateDevice(vid_device_t* device)
{
    LPDIRECTDRAW lpDD;
    HRESULT hr;

    // if the device isn't changing, skip out
	if (vid_currentDevice == device)
		return;

    // shut the current device down
	DeviceShutdown();
    
    // assign the new active device
    vid_currentDevice = device;

    // don't start a new device when null is passed in
    if (!device)
        return;
    
    // create the new device interface
    if (device->isNullGuid)
        hr = DirectDrawCreate(NULL, &lpDD, NULL);
    else
        hr = DirectDrawCreate(&device->deviceGuid, &lpDD, NULL);
    if (hr != DD_OK)
        SYS_Errorf("ActivateDevice: DirectDrawCreate on device %s(%s)",
            device->name, device->description);
    
	if (lpDD->QueryInterface(IID_IDirectDraw2, (LPVOID*)&vid_lpDD2) != DD_OK)
        SYS_Errorf("ActivateDevice: QueryInterface lpDD2 on device %s(%s)",
            device->name, device->description);
    
    lpDD->Release();
}

// activate a video mode
static void ActivateMode(vid_mode_t* mode)
{
    DDSURFACEDESC ddsdesc;
    DDSCAPS ddscaps;
    PALETTEENTRY pal[256];
    dword i;
    byte* bptr;
    PALETTEENTRY* palptr;
    HRESULT hr;

    // if the mode isn't changing, skip out
	if (vid_currentMode == mode)
		return;

	// shut the current mode down
	ModeShutdown();

    // switch devices as necessary
    if (!mode)
        ActivateDevice(NULL);
    else
        ActivateDevice(mode->device);

    // assign the new active mode
    vid_currentMode = mode;

    // don't set a new mode when null is passed in
    if (!mode)
        return;

    // change the cooperative level
    if ((hr = vid_lpDD2->SetCooperativeLevel(sysw_hwnd,
        DDSCL_FULLSCREEN|DDSCL_EXCLUSIVE|DDSCL_ALLOWREBOOT)) != DD_OK)
    {
        SYS_Errorf("ActivateMode: SetCooperativeLevel on device %s(%s)",
            mode->device->name, mode->device->description);
    }

    // change the display mode
    if (vid_lpDD2->SetDisplayMode(mode->width, mode->height,
        mode->bitDepth, 0, 0) != DD_OK)
    {
        SYS_Errorf("ActivateMode: SetDisplayMode %dx%dx%d on device %s(%s)",
            mode->width, mode->height, mode->bitDepth,
            mode->device->name, mode->device->description);
    }

    // get the primary surface
	memset(&ddsdesc, 0, sizeof(ddsdesc));
	ddsdesc.dwSize = sizeof(ddsdesc);
	ddsdesc.dwFlags = 0;
	ddsdesc.dwBackBufferCount = 1; ddsdesc.dwFlags |= DDSD_BACKBUFFERCOUNT;
	ddsdesc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_COMPLEX | DDSCAPS_FLIP; ddsdesc.dwFlags |= DDSD_CAPS;

    if (vid_lpDD2->CreateSurface(&ddsdesc, &vid_lpDDSPrimary, NULL) != DD_OK)
        SYS_Errorf("ActivateMode: CreateSurface 1 in mode %dx%dx%d on device %s(%s)",
            mode->width, mode->height, mode->bitDepth,
            mode->device->name, mode->device->description);

    // get the back buffer
    ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
    if (vid_lpDDSPrimary->GetAttachedSurface(&ddscaps, &vid_lpDDSBack) != DD_OK)
        SYS_Errorf("ActivateMode: GetAttachedSurface in mode %dx%dx%d on device %s(%s)",
            mode->width, mode->height, mode->bitDepth,
            mode->device->name, mode->device->description);

    // get the offscreen surface
    memset(&ddsdesc, 0, sizeof(ddsdesc));
    ddsdesc.dwSize = sizeof(ddsdesc);
    ddsdesc.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
    ddsdesc.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;
    ddsdesc.dwWidth = mode->width;
    ddsdesc.dwHeight = mode->height;

    if (vid_lpDD2->CreateSurface(&ddsdesc, &vid_lpDDSOffscreen, NULL) != DD_OK)
        SYS_Errorf("ActivateMode: CreateSurface 2 in mode %dx%dx%d on device %s(%s)",
            mode->width, mode->height, mode->bitDepth,
            mode->device->name, mode->device->description);

    // lock the offscreen surface and get our current pitch and buffer
    memset(&ddsdesc, 0, sizeof(ddsdesc));
    ddsdesc.dwSize = sizeof(ddsdesc);
    if (vid_lpDDSOffscreen->Lock(NULL, &ddsdesc, DDLOCK_WAIT, NULL) != DD_OK)
        SYS_Errorf("ActivateMode: Surface Lock in mode %dx%dx%d on device %s(%s)",
            mode->width, mode->height, mode->bitDepth,
            mode->device->name, mode->device->description);	
	vid_ScreenPitch = ddsdesc.lPitch;
    vid_ScreenPtr = (byte*)ddsdesc.lpSurface;
    
	// clear the offscreen surface
	for (i=0;i<mode->height;i++)
        memset(vid_ScreenPtr + i*vid_ScreenPitch, 0, vid_ScreenPitch);
    
	// get our other current values
	vid_ScreenHeight = mode->height;
    vid_ScreenWidth = mode->width;
    vid_ScreenBitDepth = mode->bitDepth;

    // plug in the palette if we're in an 8-bit mode
    if (mode->bitDepth == 8)
    {
        for (i=0,palptr=pal,bptr=vid_Palette; i<256; i++,palptr++,bptr+=3)
        {
            palptr->peRed = bptr[0];
            palptr->peGreen = bptr[1];
            palptr->peBlue = bptr[2];
            palptr->peFlags = 0;
        }
        if (vid_lpDD2->CreatePalette(DDPCAPS_8BIT | DDPCAPS_ALLOW256,
            pal, &vid_lpDDPalette, NULL) != DD_OK)
            SYS_Errorf("ActivateMode: CreatePalette in mode %dx%dx%d on device %s(%s)",
                mode->width, mode->height, mode->bitDepth,
                mode->device->name, mode->device->description);
        
        if (vid_lpDDSPrimary->SetPalette(vid_lpDDPalette) != DD_OK)
            SYS_Errorf("ActivateMode: SetPalette in mode %dx%dx%d on device %s(%s)",
                mode->width, mode->height, mode->bitDepth,
                mode->device->name, mode->device->description);
    }
}

// ddraw enumdevice callback, generates our device list
static BOOL WINAPI EnumDevCallback(GUID FAR* guidPtr,
                                   char* description,
                                   char* name,
                                   void* user)
{        
    if (vid_numDevices >= VID_MAXDEVICES)
        return(DDENUMRET_CANCEL);

    // a NULL guidPtr may be passed in for the primary device
	if (guidPtr)
    {
        vid_Devices[vid_numDevices].deviceGuid = *guidPtr;
        vid_Devices[vid_numDevices].isNullGuid = 0;
    }
    else
    {
        vid_Devices[vid_numDevices].isNullGuid = 1;
    }
    strcpy(vid_Devices[vid_numDevices].name, name);
    strcpy(vid_Devices[vid_numDevices].description, description);

    // activate the device to get capabilities
	ActivateDevice(&vid_Devices[vid_numDevices]);

    // pull in the capabilities structure
	vid_Devices[vid_numDevices].devCaps.dwSize = sizeof(DDCAPS);
    vid_Devices[vid_numDevices].helCaps.dwSize = sizeof(DDCAPS);
    if (vid_lpDD2->GetCaps(&vid_Devices[vid_numDevices].devCaps,
        &vid_Devices[vid_numDevices].helCaps) != DD_OK)
    {
        SYS_Errorf("EnumDevCallback: GetCaps on device %s(%s)",
            vid_Devices[vid_numDevices].name,
            vid_Devices[vid_numDevices].description);
    }

    // shut down the device
	ActivateDevice(NULL);

    vid_numDevices++;
    return(DDENUMRET_OK);
}

// ddraw enummode callback, generates our mode list
static BOOL WINAPI EnumModeCallback(DDSURFACEDESC* surfaceDesc,
                                    void* user)
{
    if (vid_numModes >= VID_MAXMODES)
        return(DDENUMRET_CANCEL);
    
    vid_Modes[vid_numModes].device = (vid_device_t*)user;
    vid_Modes[vid_numModes].width = surfaceDesc->dwWidth;
    vid_Modes[vid_numModes].height = surfaceDesc->dwHeight;
    vid_Modes[vid_numModes].bitDepth = surfaceDesc->ddpfPixelFormat.dwRGBBitCount;
    vid_Modes[vid_numModes].refresh = surfaceDesc->dwRefreshRate;
    vid_Modes[vid_numModes].surfDesc = *surfaceDesc;

    vid_numModes++;
    return(DDENUMRET_OK);
}

//============================================================================
//    INTERFACE FUNCTIONS
//============================================================================
void VIDW_Init()
{
    dword i;
    
    // enumerate devices
    if (DirectDrawEnumerate((LPDDENUMCALLBACK)EnumDevCallback, NULL) != DD_OK)
        SYS_Errorf("VID_Init failure: DirectDrawEnumerate");
    for (i=0;i<vid_numDevices;i++)
    {
        // activate the device
		ActivateDevice(&vid_Devices[i]);

        // set the cooperative level temporarily
		if (vid_lpDD2->SetCooperativeLevel(sysw_hwnd,
            DDSCL_FULLSCREEN|DDSCL_EXCLUSIVE|DDSCL_NOWINDOWCHANGES) != DD_OK)
            SYS_Errorf("VID_Init failure: SetCooperativeLevel on device %s(%s)",
                vid_Devices[i].name, vid_Devices[i].description);

        // enumerate modes in this device 
		if (vid_lpDD2->EnumDisplayModes(0, NULL, &vid_Devices[i],
            (LPDDENUMMODESCALLBACK)EnumModeCallback) != DD_OK)
            SYS_Errorf("VID_Init failure: EnumDisplayModes on device %s(%s",
                vid_Devices[i].name, vid_Devices[i].description);

        // change the cooperative level back
		if (vid_lpDD2->SetCooperativeLevel(sysw_hwnd,
            DDSCL_NORMAL) != DD_OK)
            SYS_Errorf("VID_Init failure: SetCooperativeLevel on device %s(%s)",
                vid_Devices[i].name, vid_Devices[i].description);

        // shut down the device
		ActivateDevice(NULL);
    }
}

void VIDW_Shutdown()
{
	// setting a null mode will free everything
	ActivateMode(NULL);
}

void VIDW_Frame()
{
    RECT r;
    HRESULT hr;
    DDSURFACEDESC ddsdesc;

	r.left = 0;
	r.right = vid_ScreenWidth;
	r.top = 0;
	r.bottom = vid_ScreenHeight;

	// we have a back buffer that we can page flip with, but for now
	// just blit from the offscreen buffer to the front buffer.
    
	// unlock the current offscreen buffer
	vid_lpDDSOffscreen->Unlock(vid_ScreenPtr);
    
	// blit it to the front
	if ((hr = vid_lpDDSPrimary->BltFast(0, 0, vid_lpDDSOffscreen, &r, DDBLTFAST_WAIT)) != DD_OK)
    {
        if (hr == DDERR_SURFACELOST)
        {
            vid_lpDDSPrimary->Restore();
            vid_lpDDSPrimary->BltFast(0, 0, vid_lpDDSOffscreen, &r, DDBLTFAST_WAIT);
        }
    }
    
	// lock the offscreen buffer again and re-get the pitch and buffer pointer
	memset(&ddsdesc, 0, sizeof(ddsdesc));
    ddsdesc.dwSize = sizeof(ddsdesc);
    vid_lpDDSOffscreen->Lock(NULL, &ddsdesc, DDLOCK_WAIT, NULL);
    vid_ScreenPitch = ddsdesc.lPitch;
    vid_ScreenPtr = (byte*)ddsdesc.lpSurface;
}

// change the video mode
bbool VIDW_SetMode(dword width, dword height, dword bitDepth)
{
    dword i;
    vid_mode_t* mode = NULL;

    // see if we have a mode that matches, don't worry about
	// the device for the moment (we'll eventually let users
	// choose the device in-game)
    for (i=0;i<vid_numModes;i++)
    {
        if ((vid_Modes[i].width == width)
         && (vid_Modes[i].height == height)
         && (vid_Modes[i].bitDepth == bitDepth))
        {
            mode = &vid_Modes[i];
            break;
        }
    }
	// if there wasn't a matching mode, cut out
    if (!mode)
        return(0);
    
	// go for it
    ActivateMode(mode);

    return(1);
}

// get the current offscreen buffer pointer
byte* VIDW_GetBuffer()
{
    return(vid_ScreenPtr);
}

// get the current screen width
dword VIDW_GetWidth()
{
    return(vid_ScreenWidth);
}

// get the current screen height
dword VIDW_GetHeight()
{
    return(vid_ScreenHeight);
}

// get the current screen pitch
// use this instead of width when calculating pixel offsets
dword VIDW_GetPitch()
{
    return(vid_ScreenPitch);
}

// copy a sequential block of color values into the palette
void VIDW_SetPaletteRange(byte* buffer, dword startColor, dword numColors)
{
    PALETTEENTRY pal[256];
    dword i;
    byte* bptr;
    PALETTEENTRY* palptr;

    // if we're not in an 8-bit video mode, abort
    if ((!vid_currentMode) || (vid_currentMode->bitDepth != 8))
        return;

    // if too many colors are being set, abort
    if ((startColor + numColors) > 256)
        return;

    // copy into our internal palette first, for when we change modes
	memcpy(&vid_Palette[startColor*3], buffer, numColors*3);

    for (i=0,palptr=pal,bptr=buffer; i<numColors; i++,palptr++,bptr+=3)
    {
        palptr->peRed = bptr[0];
        palptr->peGreen = bptr[1];
        palptr->peBlue = bptr[2];
        palptr->peFlags = 0;
    }
    vid_lpDDPalette->SetEntries(0, startColor, numColors, pal);
}

// set a single palette entry
void VIDW_SetPaletteEntry(byte color, byte r, byte g, byte b)
{
    PALETTEENTRY palentry;

    // if we're not in an 8-bit video mode, abort
    if ((!vid_currentMode) || (vid_currentMode->bitDepth != 8))
        return;

    // copy into internal palette first
	vid_Palette[color*3] = r;
    vid_Palette[color*3+1] = g;
    vid_Palette[color*3+2] = b;
    
    vid_lpDDPalette->GetEntries(0, color, 1, &palentry);
    palentry.peRed = r;
    palentry.peGreen = g;
    palentry.peBlue = b;
    vid_lpDDPalette->SetEntries(0, color, 1, &palentry);
}

// get the current colors of a palette entry
void VIDW_GetPaletteEntry(byte color, byte *r, byte *g, byte *b)
{
    // if we're not in an 8-bit video mode, abort
    if ((!vid_currentMode) || (vid_currentMode->bitDepth != 8))
        return;

    if (r)
        *r = vid_Palette[color*3];
    if (g)
        *g = vid_Palette[color*3+1];
    if (b)
        *b = vid_Palette[color*3+2];
}

//============================================================================
//    INTERFACE CLASS BODIES
//============================================================================

//****************************************************************************
//**
//**    END MODULE VID_WIN.CPP
//**
//****************************************************************************

