//****************************************************************************
//**
//**    XF_FILE.CPP
//**    Extended File System
//**
//**    Copyright (c) 1998 Chris Hargrove
//**    Notice: Usage of any code in this file is subject to the rules
//**            described in the LICENSE.TXT file included in this directory.
//**            Reading, compiling, or otherwise using this code constitutes
//**            automatic acceptance of the rules in said text file.
//**
//**    Note: Some functions like _findfirst and _findnext are not portable,
//**    but virtually every compiler on every platform has some equivalent,
//**    so change the FindFiles function as necessary.  Also, Unix-based
//**    ports would need to switch backslashes to forward slashes internally,
//**    as well as flip bigendian/littleendian values.
//**
//****************************************************************************
//============================================================================
//    IMPLEMENTATION HEADERS
//============================================================================
#include <direct.h>
#include <io.h>
#include "sys_defs.h"
#include "xf_file.h"
//============================================================================
//    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
//============================================================================
// archive definitions
#define XF_ARCHIVE_EXT	"MDA"
#define XF_ARCHIVE_MARKER ('M'+('D'<<8)+('A'<<16)+(' '<<24))
#define XF_ARCHIVE_VERSION_MAJOR 1
#define XF_ARCHIVE_VERSION_MINOR 0

// archive entry flags
enum
{
	AEF_SUBARCHIVE	= 0x01	// entry is a subdirectory archive
};

#define XF_MAXARCHIVES  256 // maximum number of registered archives
#define XF_MAXFILES     128 // maximum number of simultaneously open files

// marker to confirm that the filePos for xf_file_t's are correct in debug mode
#ifdef _DEBUG
#define VERIFY_FILEPOS(file) \
{ \
	if ((file)->filePos != (ftell((file)->pfile) - (file)->fileStart)) \
		SYS_Errorf("VERIFY_FILEPOS failure"); \
}
#else
#define VERIFY_FILEPOS(file) {}
#endif

// default file system root directory
#define XF_DEFAULTFILEROOT	"resource"

//============================================================================
//    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
//============================================================================

#pragma pack(1) // keep archive file structs packed to original spec

// archive format
typedef struct
{
	dword marker;		// XF_ARCHIVE_MARKER
	byte versionMajor;	// XF_ARCHIVE_VERSION_MAJOR
	byte versionMinor;	// XF_ARCHIVE_VERSION_MINOR
	dword arcSize;		// archive size in bytes (includes header & directory)
	dword dirOffset;	// directory offset from beginning of archive
	dword dirEntries;	// number of entries in directory
	byte reserved[14];	// reserved, set to zero
} xf_archiveHeader_t;

typedef struct
{
	char name[44];		// name of file or subarchive, null-terminated
	dword fileOfs;		// offset of file data in archive
	dword fileLen;		// file length in bytes
	byte flags;			// AEF_ flags
	byte reserved[11];	// reserved, set to zero
} xf_archiveEntry_t;

#pragma pack() // restore default packing

typedef struct xf_archive_s xf_archive_t;
struct xf_archive_s
{
	char name[SYS_MAX_PATH]; // for root archives, actual filename for archive
							 // for non-root archives, subdirectory name
	dword startOffset; // offset in the file for this archive
	xf_archiveHeader_t header; // header of archive file
	xf_archiveEntry_t* dir; // loaded directory of archive file
	xf_archive_t *firstChild, *next; // child and sibling next pointers
	xf_archive_t *rootParent; // for non-root archives, pointer to root parent
};

// file handles
typedef struct
{
	FILE* pfile;
	dword fileStart, fileLength;
	dword filePos; // current position (should always equal ftell(pfile) - fileStart)
} xf_file_t;

//============================================================================
//    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE DATA
//============================================================================
static xf_archive_t xf_Archives[XF_MAXARCHIVES]; // loaded archive pool
static xf_archive_t xf_ArchiveRoot; // fake parent to all subarchive parents
static xf_file_t xf_Files[XF_MAXFILES]; // file handle pool
static char xf_RootPath[SYS_MAX_PATH]; // root XF_Open directories
static sdword xf_FindFile_FoundHandle = -1;

//============================================================================
//    INTERFACE DATA
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
//============================================================================
static xf_archiveEntry_t* ArchiveFindEntry(xf_archive_t* parent, char* name);
static xf_archive_t *FindArchive(xf_archive_t* parent, char* name);
static xf_archive_t *RegisterArchive(xf_archive_t* parent, char* name);
static xf_file_t* FileForIndex(dword index);
static dword IndexForFile(xf_file_t* file);
static xf_file_t* AllocFile();
static char* FindFiles(char* filespec, bbool* isDir, dword* fileSize);
static void RecursiveArchiveCreate(char* inSubdir, FILE* fp);
static void RecursiveArchiveExtract(char* subdir, xf_archive_t* arc);
static dword ArchiveFileOpen(char* filename);

//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTIONS
//============================================================================

// find an entry in an archive directory
static xf_archiveEntry_t* ArchiveFindEntry(xf_archive_t* parent, char* name)
{
	dword i;
	for (i=0;i<parent->header.dirEntries;i++)
	{
		if (!stricmp(parent->dir[i].name, name))
			return(&parent->dir[i]);
	}
	return(NULL);
}

// find an existing archive under a given parent
static xf_archive_t *FindArchive(xf_archive_t* parent, char* name)
{
	xf_archive_t *child;

	if (!parent)
		parent = &xf_ArchiveRoot;
	for (child=parent->firstChild; child; child=child->next)
	{
		if (!stricmp(child->name, name))
			return(child);
	}
	return(NULL);
}

// register (find or add) an archive to the loaded archive tree
// if parent is NULL, then name is the filename (no ext) of an archive file
// if parent is non-NULL, then name is a subdirectory name under the parent
static xf_archive_t *RegisterArchive(xf_archive_t* parent, char* name)
{
	static dword arcindex = 0;
	xf_archive_t* out;
	xf_archiveEntry_t* entry;
	char namebuf[SYS_MAX_PATH];
	FILE* fp;

	if (!name)
		return(NULL);
	strcpy(namebuf, name);
	if (!parent)
	{		
		M_SuggestFileExtension(namebuf, XF_ARCHIVE_EXT);
		parent = &xf_ArchiveRoot;
		if (out = FindArchive(parent, namebuf))
			return(out); // archive is already loaded
		// load a new archive file
		fp = fopen(namebuf, "rb");
		if (!fp)
			return(NULL);
	}
	else
	{
		if (out = FindArchive(parent, namebuf))
			return(out); // archive is already loaded
		// find the appropriate subdirectory
		entry = ArchiveFindEntry(parent, namebuf);
		if (!entry)
			return(NULL); // no such subdirectory
		if (!(entry->flags & AEF_SUBARCHIVE))
			return(NULL); // entry is not a subdirectory
		fp = fopen(parent->rootParent->name, "rb");
		if (!fp)
			return(NULL);
		fseek(fp, parent->startOffset + entry->fileOfs, SEEK_SET);
	}

	if (arcindex >= XF_MAXARCHIVES)
	{
		// no more archive nodes available
		fclose(fp);
		return(NULL);
	}
	out = &xf_Archives[arcindex];
	out->startOffset = ftell(fp);
	if (fread(&out->header, sizeof(out->header), 1, fp) != 1)
	{
		fclose(fp);
		return(NULL);
	}
	if ((out->header.marker != XF_ARCHIVE_MARKER)
	 || (out->header.versionMajor != XF_ARCHIVE_VERSION_MAJOR)
	 || (out->header.versionMinor != XF_ARCHIVE_VERSION_MINOR))
	{
		// incorrect version
		fclose(fp);
		return(NULL);
	}
	out->dir = NULL;
	if (out->header.dirEntries)
	{
		out->dir = SYS_MALLOC(xf_archiveEntry_t, out->header.dirEntries);
		fseek(fp, out->startOffset + out->header.dirOffset, SEEK_SET);
		if (fread(out->dir, sizeof(xf_archiveEntry_t), out->header.dirEntries, fp)
			!= out->header.dirEntries)
		{
			SYS_FREE(out->dir);
			fclose(fp);
			return(NULL);
		}
	}
	// archive loaded
	arcindex++;
	out->firstChild = NULL;
	out->next = parent->firstChild;
	parent->firstChild = out;
	if (parent == &xf_ArchiveRoot)
	{
		_getcwd(out->name, SYS_MAX_PATH);
		strcat(out->name, "\\");
		strcat(out->name, namebuf);
		out->rootParent = out;
	}
	else
	{
		strcpy(out->name, namebuf);
		out->rootParent = parent->rootParent;
	}
	return(out);
}

// get actual file handle for index
static xf_file_t* FileForIndex(dword index)
{
	if ((!index) || (index > XF_MAXFILES))
		return(NULL);
	return(&xf_Files[index-1]);
}

// get index for a file handle
static dword IndexForFile(xf_file_t* file)
{
	if (!file)
		return(NULL);
	return((dword)(file - xf_Files)+1);
}

// allocate an available file handle
static xf_file_t* AllocFile()
{
	dword i;
	for (i=0;i<XF_MAXFILES;i++)
	{
		if (!xf_Files[i].pfile)
			return(&xf_Files[i]);
	}
	return(NULL);
}

// search for actual files
// note that this function probably isn't directly portable
static char* FindFiles(char* filespec, bbool* isDir, dword* fileSize)
{
	static _finddata_t data;
	if (filespec)
	{
        if ((xf_FindFile_FoundHandle = _findfirst(filespec, &data)) == -1)
			return(NULL);
	}
	else
	{
        if (xf_FindFile_FoundHandle == -1)
			return(NULL);
        if (_findnext(xf_FindFile_FoundHandle, &data) == -1)
		{
            xf_FindFile_FoundHandle = -1;
			return(NULL);
		}
	}
	if (isDir)
		*isDir = ((data.attrib & _A_SUBDIR)!=0);
	if (fileSize)
		*fileSize = data.size;
	return(data.name);
}

// called recursively by XF_ArchiveCreate
static void RecursiveArchiveCreate(char* inSubdir, FILE* fp)
{
	static char* startspec = "*.*";
	char oldpath[SYS_MAX_PATH], subdir[SYS_MAX_PATH];
	dword startPos;
	dword numEntries, size;
	xf_archiveHeader_t header;
	xf_archiveEntry_t *dir;
	char *spec, *specfile;
	bbool isDir;
	FILE* datafile;
	byte* databuf;
	sdword foundhandleBackup;

	// backup current directory and file position
	_getcwd(oldpath, SYS_MAX_PATH);
	startPos = ftell(fp);

	// change to subdirectory
	strcpy(subdir, inSubdir);
	_chdir(subdir);
	
	// setup archive header
	header.marker = XF_ARCHIVE_MARKER;
	header.versionMajor = XF_ARCHIVE_VERSION_MAJOR;
	header.versionMinor = XF_ARCHIVE_VERSION_MINOR;
	memset(header.reserved, 0, sizeof(header.reserved));
	// fill in the rest after everything's written
	header.arcSize = 0;
	header.dirOffset = 0;
	header.dirEntries = 0;
	
	// write out initial header
	if (fwrite(&header, sizeof(header), 1, fp) != 1)
	{
		SYS_Errorf("XF_ArchiveCreate: Unable to write header for %s", subdir);
	}

	// count number of entries archive will need
	numEntries = 0;
	spec = startspec;
	while (specfile = FindFiles(spec, &isDir, NULL))
	{
		spec = NULL;
		if (isDir)
		{
			// don't count "." or ".." directories
			if ((!strcmp(specfile, ".")) || (!strcmp(specfile, "..")))
				continue;
		}
		numEntries++;
	}

	// allocate directory
	dir = SYS_MALLOC(xf_archiveEntry_t, numEntries);
	memset(dir, 0, numEntries*sizeof(xf_archiveEntry_t));
	header.dirEntries = numEntries;

	// add all regular files to directory
	numEntries = 0;
	spec = startspec;
	while (specfile = FindFiles(spec, &isDir, &size))
	{
		spec = NULL;
		if (isDir)
			continue;
		datafile = fopen(specfile, "rb");
		if (!datafile)
		{
			SYS_Debugf("Warning: Unable to open %s", specfile);
			continue;
		}
		databuf = SYS_MALLOC(byte, size);
		if (fread(databuf, 1, size, datafile) != size)
		{
			SYS_Debugf("Warning: Unable to read all %d bytes of %s, skipping file", size, specfile);
			SYS_FREE(databuf);
			continue;
		}
		fclose(datafile);
		strcpy(dir[numEntries].name, specfile);
		dir[numEntries].fileOfs = ftell(fp) - startPos;
		SYS_Debugf("Adding %s\\%s\\%s...\n", strupr(oldpath), strupr(subdir), strupr(specfile));
		if (fwrite(databuf, 1, size, fp) != size)
		{
			SYS_Debugf("Warning: Unable to write all %d bytes of %s to archive", size, specfile);
		}
		dir[numEntries].fileLen = (ftell(fp) - startPos) - dir[numEntries].fileOfs;
		SYS_FREE(databuf);
		numEntries++;
	}
	
	// add all subdirectories and recurse to them
	spec = startspec;
	while (specfile = FindFiles(spec, &isDir, &size))
	{
		spec = NULL;
		if (!isDir)
			continue;
		// don't count "." or ".." directories
		if ((!strcmp(specfile, ".")) || (!strcmp(specfile, "..")))
			continue;
		strcpy(dir[numEntries].name, specfile);
		dir[numEntries].fileOfs = ftell(fp) - startPos;
		
		//SYS_Debugf("Recursing to %s\\%s\\%s...\n", strupr(oldpath), strupr(subdir), strupr(specfile));
        foundhandleBackup = xf_FindFile_FoundHandle;
		RecursiveArchiveCreate(specfile, fp);
        xf_FindFile_FoundHandle = foundhandleBackup;

		dir[numEntries].fileLen = (ftell(fp) - startPos) - dir[numEntries].fileOfs;
		dir[numEntries].flags = AEF_SUBARCHIVE;
		numEntries++;
	}

	// write out directory
	if (numEntries != header.dirEntries)
		SYS_Errorf("XF_ArchiveCreate: Directory entry count out of sync for %s", subdir);
	header.dirOffset = ftell(fp) - startPos;
	if (fwrite(dir, sizeof(xf_archiveEntry_t), header.dirEntries, fp)
		!= header.dirEntries)
	{
		// this one can't be just a warning, since a corrupted directory is useless
		SYS_Errorf("XF_ArchiveCreate: Unable to write directory for %s", subdir);
	}
	
	// rewrite header
	header.arcSize = ftell(fp) - startPos;
	fseek(fp, startPos, SEEK_SET);
	if (fwrite(&header, sizeof(header), 1, fp) != 1)
	{
		SYS_Errorf("XF_ArchiveCreate: Unable to write header for %s", subdir);
	}
	fseek(fp, startPos + header.arcSize, SEEK_SET);

	// return to backup directory
	_chdir(oldpath);
}

// called recursively by XF_ArchiveExtract
static void RecursiveArchiveExtract(char* subdir, xf_archive_t* arc)
{
	dword i;
	char oldpath[SYS_MAX_PATH];
	FILE *fp, *datafile;
	byte *databuf;
	
	if ((!subdir) || (!arc))
		return;
	_getcwd(oldpath, SYS_MAX_PATH);
	SYS_Debugf("Creating %s\\%s\\\n", oldpath, subdir);
	_mkdir(subdir);
	_chdir(subdir);
	// extract files
	for (i=0;i<arc->header.dirEntries;i++)
	{
		if (arc->dir[i].flags & AEF_SUBARCHIVE)
			continue;
		SYS_Debugf("Extracting %s\\%s\\%s\n", oldpath, subdir, arc->dir[i].name);
		fp = fopen(arc->rootParent->name, "rb");
		if (!fp)
		{
			SYS_Debugf("XF_ArchiveExtract: Unable to open archive file\n");
			continue;
		}
		datafile = fopen(arc->dir[i].name, "wb");
		if (!datafile)
		{
			SYS_Debugf("XF_ArchiveExtract: Unable to create target file\n");
			fclose(fp);
			continue;
		}
		fseek(fp, arc->dir[i].fileOfs + arc->startOffset, SEEK_SET);
		databuf = SYS_MALLOC(byte, arc->dir[i].fileLen);
		fread(databuf, 1, arc->dir[i].fileLen, fp);
		fclose(fp);
		fwrite(databuf, 1, arc->dir[i].fileLen, datafile);
		fclose(datafile);
		SYS_FREE(databuf);
	}
	// recurse to subdirs
	for (i=0;i<arc->header.dirEntries;i++)
	{
		if (arc->dir[i].flags & AEF_SUBARCHIVE)
			RecursiveArchiveExtract(arc->dir[i].name, RegisterArchive(arc, arc->dir[i].name));
	}
	_chdir(oldpath);
}

// open a file, first checking if an actual file exists,
// then checking any possible archives
static dword ArchiveFileOpen(char* filename)
{
	xf_file_t *fp;
	xf_archive_t *arc;
	char buf[SYS_MAX_PATH];
	char *ptr, *dirname;
	xf_archiveEntry_t* entry;

	if (!filename)
		return(0);
	fp = AllocFile();
	if (!fp)
		return(0);
	
	// try actual file first
	fp->pfile = fopen(filename, "rb");
	if (fp->pfile)
	{
		fp->fileStart = 0;
		fseek(fp->pfile, 0, SEEK_END);
		fp->fileLength = ftell(fp->pfile);
		fseek(fp->pfile, 0, SEEK_SET);
		fp->filePos = 0;
		return(IndexForFile(fp));
	}
	
	// file wasn't there, check the archives
	strcpy(buf, filename);
	dirname = buf;
	arc = NULL;
	while (ptr = strchr(dirname, '\\'))
	{
		*ptr = 0;
		arc = RegisterArchive(arc, dirname);
		if (!arc)
			return(0); // no matching archive
		dirname = ptr+1;
	}
	entry = ArchiveFindEntry(arc, dirname);
	if (!entry)
		return(0); // no entry for this file
	fp->pfile = fopen(arc->rootParent->name, "rb");
	if (!fp->pfile)
		return(0); // unable to open archive file
	fp->fileStart = arc->startOffset + entry->fileOfs;
	fseek(fp->pfile, fp->fileStart, SEEK_SET);
	fp->fileLength = entry->fileLen;
	fp->filePos = 0;
	return(IndexForFile(fp));
}

//============================================================================
//    INTERFACE FUNCTIONS
//============================================================================

// initialization/shutdown

void XF_Init()
{
	byte parm;
	char rootpath[SYS_MAX_PATH];
	
	if (parm = M_CmdLineOption("packcreate", 1))
	{
		// generate a pack archive for a distribution subdirectory
		SYS_InitDebug();
		if (!XF_ArchiveCreate(SYS_Argv(parm+1)))
			SYS_Errorf("Unable to create archive for %s", SYS_Argv(parm+1));
		SYS_Alertf("Packing complete.");
		SYS_ShutdownDebug();
		SYS_Quit();
	}
	if (parm = M_CmdLineOption("packextract", 1))
	{
		// extract a pack archive in a distribution subdirectory
		SYS_InitDebug();
		if (!XF_ArchiveExtract(SYS_Argv(parm+1)))
			SYS_Errorf("Unable to extract archive %s", SYS_Argv(parm+1));
		SYS_Alertf("Extraction complete.");
		SYS_ShutdownDebug();
		SYS_Quit();
	}
	
	// add one or more directories to the rootpath
	rootpath[0] = 0;
	if (parm = M_CmdLineOption("rootpath", 1))
	{
		strcpy(rootpath, SYS_Argv(parm+1));
		strcat(rootpath, ";");
	}
	strcat(rootpath, XF_DEFAULTFILEROOT); // add on default rootpath
	XF_RootPath(rootpath);
}

void XF_Shutdown()
{
	dword i;

    for (i=0;i<XF_MAXFILES;i++)
	{
		if (xf_Files[i].pfile)
	        fclose(xf_Files[i].pfile);
	}
}

// standard file operations

// open a file for reading, returns file handle or zero on failure
// pure files are checked first, followed by archived files
// checks all rootpath directories in order
dword XF_Open(char* filename)
{
	char* ptr = xf_RootPath;
	char filebuf[SYS_MAX_PATH];
	dword result;

	// walk path directories, plus an additional check of current directory
	do
	{
		filebuf[0] = 0;
		if (*ptr)
		{
			strcpy(filebuf, ptr);
			strcat(filebuf, "\\");
		}
		strcat(filebuf, filename);
		if (result = ArchiveFileOpen(filebuf))
			return(result);
		ptr += strlen(ptr)+1; // go to next string
	} while (*ptr);
	return(0);
}

// create a file for writing
dword XF_Create(char* filename)
{
	xf_file_t *fp;

	if (!filename)
		return(0);
	fp = AllocFile();
	if (!fp)
		return(0);
	fp->pfile = fopen(filename, "wb");
	if (!fp->pfile)
		return(0);
	fp->fileStart = fp->fileLength = fp->filePos = 0;

	return(IndexForFile(fp));
}

// close a file
bbool XF_Close(dword fileHandle)
{
	xf_file_t *fp = FileForIndex(fileHandle);
	if ((!fp) || (!fp->pfile))
		return(0);
	fclose(fp->pfile);
	fp->pfile = NULL;
	return(1);
}

// read from file
bbool XF_Read(void* buf, dword size, dword fileHandle)
{
	xf_file_t *fp = FileForIndex(fileHandle);
	if ((!fp) || (!fp->pfile) || (!size))
		return(0);
	if ((fp->filePos + size) > fp->fileLength)
		return(0); // would exceed file length
	if (fread(buf, 1, size, fp->pfile) != size)
	{
		// couldn't read everything, restore old position
		if (fseek(fp->pfile, fp->fileStart + fp->filePos, SEEK_SET))
			SYS_Errorf("XF_Read Error: Could not re-seek");
		return(0);
	}
	fp->filePos += size;
	VERIFY_FILEPOS(fp);
	return(1);
}

// write to file
bbool XF_Write(void* buf, dword size, dword fileHandle)
{
	xf_file_t *fp = FileForIndex(fileHandle);
	if ((!fp) || (!fp->pfile) || (!size))
		return(0);
	if (fwrite(buf, 1, size, fp->pfile) != size)
	{
		// couldn't write everything, restore old position
		if (fseek(fp->pfile, fp->fileStart + fp->filePos, SEEK_SET))
			SYS_Errorf("XF_Write Error: Could not re-seek");
		return(0);
	}
	fp->filePos += size;
	if (fp->filePos > fp->fileLength)
		fp->fileLength = fp->filePos;
	VERIFY_FILEPOS(fp);
	return(1);
}

// seek to absolute position in file
bbool XF_Seek(dword fileHandle, dword position)
{
	xf_file_t *fp = FileForIndex(fileHandle);
	if ((!fp) || (!fp->pfile))
		return(0);
	if (position > fp->fileLength)
		return(0); // can't seek past end of file
	if (fseek(fp->pfile, fp->fileStart + position, SEEK_SET))
		return(0); // if fseek returns nonzero it failed
	fp->filePos = position;
	VERIFY_FILEPOS(fp);
	return(1);
}

// seek to position in file relative to current position
bbool XF_SeekCurrent(dword fileHandle, sdword position)
{
	xf_file_t *fp = FileForIndex(fileHandle);
	sdword targetPos;
	if ((!fp) || (!fp->pfile))
		return(0);
	targetPos = (sdword)fp->filePos + position;
	if ((targetPos < 0) || (targetPos > (sdword)fp->fileLength))
		return(0); // can't seek past end of file
	if (fseek(fp->pfile, fp->fileStart + targetPos, SEEK_SET))
		return(0); // if fseek returns nonzero it failed
	fp->filePos = targetPos;
	VERIFY_FILEPOS(fp);
	return(1);
}

// seek backwards from the end of the file
bbool XF_SeekEnd(dword fileHandle, dword position)
{
	xf_file_t *fp = FileForIndex(fileHandle);
	sdword targetPos;
	if ((!fp) || (!fp->pfile))
		return(0);
	targetPos = (sdword)fp->fileLength - position;
	if ((targetPos < 0) || (targetPos > (sdword)fp->fileLength))
		return(0); // can't seek past end of file
	if (fseek(fp->pfile, fp->fileStart + targetPos, SEEK_SET))
		return(0); // if fseek returns nonzero it failed
	fp->filePos = targetPos;
	VERIFY_FILEPOS(fp);
	return(1);
}

// return current position in file
dword XF_Tell(dword fileHandle)
{
	xf_file_t *fp = FileForIndex(fileHandle);
	if ((!fp) || (!fp->pfile))
		return(0);
	VERIFY_FILEPOS(fp);
	return(fp->filePos);
}

// return size of file
dword XF_Size(dword fileHandle)
{
	xf_file_t *fp = FileForIndex(fileHandle);
	if ((!fp) || (!fp->pfile))
		return(0);
	return(fp->fileLength);
}

// archive operations

// create an archive file for a subdirectory
// must be directly underneath distribution directory (no slashes)
bbool XF_ArchiveCreate(char* subdir)
{
	char filename[SYS_MAX_PATH];
	FILE *fp;

	if (strchr(subdir, '\\') || strchr(subdir, ':'))
		return(0); // no slashes or absolute drives allowed
	strcpy(filename, subdir);
	M_SuggestFileExtension(filename, XF_ARCHIVE_EXT);
	fp = fopen(filename, "wb");
	if (!fp)
		return(0);

	RecursiveArchiveCreate(subdir, fp);

	fclose(fp);	
	return(1);
}

// create a subdirectory for a particular archive name and extract
// must be directly underneath distribution directory (no slashes)
bbool XF_ArchiveExtract(char* subdir)
{
	char filename[SYS_MAX_PATH];
	xf_archive_t* arc;

	if (strchr(subdir, '\\') || strchr(subdir, ':'))
		return(0); // no slashes or absolute drives allowed
	strcpy(filename, subdir);
	M_SuggestFileExtension(filename, XF_ARCHIVE_EXT);

	arc = RegisterArchive(NULL, filename);
	if (!arc)
		return(0);

	RecursiveArchiveExtract(subdir, arc);

	return(1);
}

// miscellaneous

// set XF_Open root dirs, separated by semicolons
bbool XF_RootPath(char* path)
{
	char* ptr;
	dword len;

	if (!path)
		return(0);
	// set xf_RootPath to the path dir strings followed by additional null
	strcpy(xf_RootPath, path);
	len = strlen(xf_RootPath);
	xf_RootPath[len+1] = 0; // set the final null
	
	// replace all semicolons will nulls
	for (ptr=xf_RootPath; ptr=strchr(ptr, ';'); ptr++)
		*ptr = 0;
	return(1);
}

//============================================================================
//    INTERFACE CLASS BODIES
//============================================================================

//****************************************************************************
//**
//**    END MODULE XF_FILE.CPP
//**
//****************************************************************************

